# Authors:
#   Keith Forsyth and Dan Kuester

import os
import time
import labbench as lb
from labbench import paramattr as attr
import typing
import typing_extensions
from typing import Union, Literal

if typing.TYPE_CHECKING:
    import pandas as pd
    import numpy as np
else:
    # delayed import for speed
    pd = lb.util.lazy_import('pandas')
    np = lb.util.lazy_import('numpy')

DataFrameType: typing_extensions.TypeAlias = 'pd.DataFrame'
SeriesType: typing_extensions.TypeAlias = 'pd.Series'
NumpyArrayType: typing_extensions.TypeAlias = 'np.ndarray'

__all__ = [
    'KeysightN9951B',
    'RohdeSchwarzFSW26SpectrumAnalyzer',
    'RohdeSchwarzFSW26IQAnalyzer',
    'RohdeSchwarzFSW26LTEAnalyzer',
    'RohdeSchwarzFSW26RealTime',
    'RohdeSchwarzFSW43Base',
    'RohdeSchwarzFSW43SpectrumAnalyzer',
    'RohdeSchwarzFSW43IQAnalyzer',
    'RohdeSchwarzFSW43LTEAnalyzer',
    'RohdeSchwarzFSW43RealTime',
]

DEFAULT_CHANNEL_NAME = 'remote'


@attr.visa_keying(remap={False: '0', True: '1'})
class KeysightN9951B(lb.VISADevice):
    """A Keysight N9951B "Field Fox".

    Attributes:
        frequency_start (int): the sweep start
    """

    frequency_start = attr.property.float(
        key='FREQ:START', min=1e6, max=43.99e9, label='Hz'
    )
    frequency_stop = attr.property.float(
        key='FREQ:STOP', min=10e6, max=44e9, label='Hz'
    )

    # the bounds on these two vary dynamically with the value of the other.
    # sets=False here to avoid accidents - use frequency_start and frequency_stop
    frequency_span = attr.property.float(key='FREQ:SPAN', sets=False, min=2, max=44e9, label='Hz')
    frequency_center = attr.property.float(
        key='FREQ:CENT', sets=False, min=2, max=26.5e9, step=1e-9, label='Hz'
    )

    initiate_continuous = attr.property.bool(key='INIT:CONT')
    reference_level = attr.property.float(
        key='DISP:WIND:TRAC1:Y:RLEV', step=1e-3, label='dB'
    )

    resolution_bandwidth = attr.property.float(
        key='BAND', min=1e3, max=5.76e6, label='Hz'
    )

    def fetch_trace(self, trace: int = 1) -> DataFrameType:
        """Get trace x values and y values using XVAL? and DATA?

        Arguments:
            trace: which trace to pull from the fieldfox
        """

        x_data = self.query(f'TRAC{trace}:XVAL?').split(',')
        y_data = self.query(f'TRAC{trace}:DATA?').split(',')
        return pd.DataFrame({'Frequency': x_data, 'Power': y_data})

    def get_marker_power(self, marker: int) -> float:
        """Get marker measurement value (on vertical axis)

        Arguments:
            marker: number on instrument display
        Returns:
            position of the marker, in units of the horizontal axis
        """
        return float(self.query(f'CALC:MARK{marker}:Y?'))

    def get_marker_position(self, marker: int) -> float:
        """Get marker position (on horizontal axis)
        Arguments:
            marker: marker number on instrument display

        Returns:
            position of the marker, in units of the horizontal axis
        """
        return float(self.query(f'CALC:MARK{marker}:X?'))

    def set_marker_position(self, marker: int, position: float):
        """Get marker position (on horizontal axis)

        Arguments:
            marker: marker number on instrument display
            position: position of the marker, in units of the horizontal axis
        """
        return self.write(f'CALC:MARK{marker}:X {position}')


@attr.method_kwarg.int('trace', min=1, max=6, help='trace index for analysis')
@attr.method_kwarg.int(
    'output_trigger_index', min=1, max=3, help='output trigger port index'
)
class RohdeSchwarzFSWBase(lb.VISADevice):
    _DATA_FORMATS = 'ASC,0', 'REAL,32', 'REAL,64', 'REAL,16'
    _CHANNEL_TYPES = 'SAN', 'IQ', 'RTIM', DEFAULT_CHANNEL_NAME
    _TRIGGER_OUT_TYPES = 'DEV', 'TARM', 'UDEF'
    _TRIGGER_DIRECTIONS = 'INP', 'OUTP'
    _CHANNEL_TYPES = None, 'SAN', 'IQ', 'RTIM'
    _CACHE_DIR = r'c:\temp\remote-cache'

    expected_channel_type: str = attr.value.str(
        None,
        allow_none=True,
        only=_CHANNEL_TYPES,
        sets=False,
        cache=True,
        help='which channel type to use',
    )
    default_window: str = attr.value.str(
        '', cache=True, help='data window number to use if unspecified'
    )
    default_trace: str = attr.value.str(
        '', cache=True, help='data trace number to use if unspecified'
    )

    # Set these in subclasses for specific FSW instruments
    frequency_center = attr.property.float(
        key='FREQ:CENT', min=0, step=1e-9, label='Hz'
    )
    frequency_span = attr.property.float(key='FREQ:SPAN', min=0, step=1e-9, label='Hz')
    frequency_start = attr.property.float(
        key='FREQ:START', min=0, step=1e-9, label='Hz'
    )
    frequency_stop = attr.property.float(key='FREQ:STOP', min=0, step=1e-9, label='Hz')
    resolution_bandwidth = attr.property.float(key='BAND', min=0, label='Hz')

    sweep_time = attr.property.float(key='SWE:TIME', label='Hz')
    sweep_time_window2 = attr.property.float(key='SENS2:SWE:TIME', label='Hz')

    initiate_continuous = attr.property.bool(key='INIT:CONT')

    reference_level = attr.method.float(
        key='DISP:TRAC{trace}:Y:RLEV', step=1e-3, label='dB'
    )

    amplitude_offset = attr.method.float(
        key='DISP:TRAC{trace}:Y:RLEV:OFFS', step=1e-3, label='dB'
    )

    output_trigger_direction = attr.method.str(
        key='OUTP:TRIG{output_trigger_index}:DIR', only=_TRIGGER_DIRECTIONS, case=False
    )
    output_trigger_type = attr.method.str(
        key='OUTP:TRIG{output_trigger_index}:OTYP', only=_TRIGGER_OUT_TYPES, case=False
    )

    input_preamplifier_enabled = attr.property.bool(key='INP:GAIN:STATE')
    input_attenuation_auto = attr.property.bool(key='INP:ATT:AUTO')
    input_attenuation = attr.property.float(key='INP:ATT', step=1, min=0, max=79)

    channel_type = attr.property.str(key='INST', only=_CHANNEL_TYPES, case=False)
    format = attr.property.str(key='FORM', only=_DATA_FORMATS, case=False)
    sweep_points = attr.property.int(key='SWE:POIN', min=1, max=100001)

    display_update = attr.property.bool(key='SYST:DISP:UPD')
    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    def verify_channel_type(self):
        valid = self.expected_channel_type, DEFAULT_CHANNEL_NAME
        if self.expected_channel_type is not None and self.channel_type not in valid:
            self._logger.warning(
                f'expected {self.expected_channel_type} mode, but got {self.channel_type}'
            )

    def open(self):
        self.format = 'REAL,32'

    def acquire_spectrogram(self, acquisition_time_sec):
        t0 = time.time()

        specs = []

        time_remaining = acquisition_time_sec
        active_time = 0

        while active_time == 0:
            # Setup
            self.clear_spectrogram()
            self.wait()
            # Give the power sensor time to arm
            lb.sleep(0.1)

            t0_active = time.time()

            # Try to trigger; block until timeout.
            with self.overlap_and_block(timeout=int(1e3 * time_remaining)):
                self.trigger_single(wait=False)
            active_time += time.time() - t0_active

            self.backend.timeout = 50000
            single = self.fetch_spectrogram(timeout=50000, timestamps='fast')
            self.backend.timeout = 1000

            if single is not None:
                specs.append(single)

            time_remaining = acquisition_time_sec - (time.time() - t0)

        specs = pd.concat(specs, axis=0) if specs else pd.DataFrame()
        return {
            'sa_spectrogram': specs,
            'sa_spectrogram_acquisition_time': time.time() - t0,
            'sa_spectrogram_active_time': active_time,
        }

    def close(self):
        try:
            self.abort()
        except BaseException:
            pass
        try:
            self.clear_status()
        except BaseException:
            pass

    def clear_status(self):
        self.write('*CLS')

    def status_preset(self):
        self.write('STAT:PRES')

    def save_state(self, name: str, basedir: Union[str, None] = None):
        """Save current state of the device to the default directory in the instrument.

        Arguments:
            path: state file location on the instrument
        """
        if basedir is None:
            path = name
        else:
            self.mkdir(basedir)
            path = basedir + '\\' + name

        self.write(f"MMEMory:STORe:STATe 1,'{path}'")
        self.wait()

    def load_state(self, name: str, basedir: Union[str, None] = None):
        """Loads a previously saved state file in the instrument

        Arguments:
            path: state file location on the instrument
        """
        if basedir is not None:
            path = basedir + '\\' + name
        if not path.endswith('.dfl'):
            path = path + '.dfl'

        if self.file_info(path) is None:
            raise FileNotFoundError(
                f'there is no file to load on the instrument at path "{path}"'
            )

        self.write(f"MMEM:LOAD:STAT 1,'{path}'")
        self.wait()

    def load_cache(self):
        cache_name = lb.util.hash_caller(2)

        try:
            self.load_state(cache_name, self._CACHE_DIR)
        except FileNotFoundError:
            return False
        else:
            self._logger.debug('Successfully loaded cached save file')
            return True

    def save_cache(self):
        cache_name = lb.util.hash_caller(2)
        self.save_state(cache_name, self._CACHE_DIR)

    def mkdir(self, path, recursive=True):
        """Make a new directory (optionally recursively) on the instrument
        if we haven't tried to make it already.
        """
        try:
            if path in self.__prev_dirs:
                return
        except AttributeError:
            self.__prev_dirs = set()

        if recursive:
            subs = path.replace('/', '\\').split('\\')
            for i in range(1, len(subs) + 1):
                self.mkdir('\\'.join(subs[:i]), recursive=False)
        else:
            with self.overlap_and_block():
                self.write(f"MMEM:MDIR '{path}'")
            self.__prev_dirs.add(path)
        return path

    def file_info(self, path):
        with self.suppress_timeout(), self.overlap_and_block(timeout=0.1):
            ret = None
            ret = self.query(f"MMEM:CAT? '{path}'")
        return ret

    def remove_window(self, name):
        self.write(f"LAY:REM '{name}'")

    def trigger_single(self, wait=True, disable_continuous=True):
        """Trigger once."""
        if disable_continuous:
            self.initiate_continuous = False
        self.write('INIT')
        if wait:
            self.wait()

    def autolevel(self):
        """Try to automatically set the reference level on the instrument, which sets the
        internal attenuation and preamplifier enable settings.
        """
        self.write('ADJ:LEV')

    def abort(self):
        self.write('ABORT')

    def apply_channel_type(self, type_=None):
        """setup a channel with name DEFAULT_CHANNEL_NAME, that has measurement type self.channel_type"""
        channel_list = self.query('INST:LIST?').replace("'", '').split(',')[1::2]
        if DEFAULT_CHANNEL_NAME in channel_list:
            self.write(
                f"INST:CRE:REPL '{DEFAULT_CHANNEL_NAME}',"
                f"{self.channel_type},'{DEFAULT_CHANNEL_NAME}'"
            )
        else:
            self.write(f"INST:CRE {self.channel_type},'{DEFAULT_CHANNEL_NAME}'")

    def channel_preset(self):
        self.write('SYST:PRES:CHAN')

    def query_ieee_array(self, msg: str) -> NumpyArrayType:
        """An alternative to self.backend.query_binary_values for fetching block data. This
        implementation works around slowness between pyvisa and the instrument that seems to
        result from transferring in chunks of size self.backend.chunk_size as implemented
        in pyvisa.

        The performance of a transfer of a spectrogram with 100,000 time samples is impacted as follows:

        # The pyvisa implementation
        >>>  %timeit -n 1 -r 1 sa.backend.query_binary_values('TRAC2:DATA? SPEC', container=np.array)
        (takes at least 10 minutes)

        # This implementation
        >>> %timeit -n 1 -r 1 sa.query_ieee_array('TRAC2:DATA? SPEC')
        (~23 sec)

        Arguments:
            msg: The SCPI command to send
        :return: a numpy array containing the response.
        """

        from pyvisa.constants import VI_SUCCESS_DEV_NPRESENT, VI_SUCCESS_MAX_CNT

        self._logger.debug(f'query {msg}')

        # The read_termination seems to cause unwanted behavior in self.backend.visalib.read
        self.backend.read_termination, old_read_term = (
            None,
            self.backend.read_termination,
        )
        self.backend.write(msg)

        try:
            with self.backend.ignore_warning(
                VI_SUCCESS_DEV_NPRESENT, VI_SUCCESS_MAX_CNT
            ):
                # Reproduce the behavior of pyvisa.util.from_ieee_block without
                # a priori access to the entire buffer.
                raw, _ = self.backend.visalib.read(self.backend.session, 2)
                digits = int(raw.decode('ascii')[1])
                raw, _ = self.backend.visalib.read(self.backend.session, digits)
                data_size = int(raw.decode('ascii'))

                # Read the actual data
                raw, _ = self.backend.visalib.read(self.backend.session, data_size)

                # Read termination characters so that the instrument doesn't show
                # a "QUERY INTERRUPTED" error when there is unread buffer
                self.backend.visalib.read(self.backend.session, len(old_read_term))
        finally:
            self.backend.read_termination = old_read_term

        data = np.frombuffer(raw, np.float32)
        self._logger.debug('      -> {} bytes ({} values)'.format(data_size, data.size))
        return data

    def fetch_horizontal(self, window=None, trace: int = None):
        if window is None:
            window = self.default_window
        if trace is None:
            trace = self.default_trace

        return self.query_ieee_array(f'TRAC{window}:DATA:X? TRACE{trace}')

    def fetch_trace(self, trace=None, horizontal=False, window=None) -> DataFrameType:
        """Fetch trace data with 'TRAC:DATA TRACE?' and return the result in
        a pandas series.fetch and return the current trace data. This does not
        initiate a trigger; this must be done separately if desired.
        (see :method:`trigger_single`). This method is meant to be used
        in all signal analyzer modes (spectrum analyzer, IQ analyzer, LTE analyzer,
        etc.), because variants of TRAC:DATA TRACE? are implemented for each.

        The `trace` and `window` parameters have different meaning in different
        modes. Specify `window` only in signal analyzer modes that support it,
        which include LTE.  When `window` is not supported, specifying it (any
        value besides `None`) will cause a timeout.

        Trace data is returned in single-precision (32-bit) binary blocks.

        If necessary, we can read the count, set count to 1, read, adjust level
        then set the count back, and read again like this:

        ::
            count = inst.query('SENSE:SWEEP:COUNT?')
            self.write('SENSE:SWEEP:COUNT 1')

        Arguments:
            trace: The trace number to query (or None, the default, to use self.default_trace)
            horizontal: Set the index of the returned Series by a call to :method:`fetch_horizontal`
            window: The window number to query (or None, the default, to use self.default_window)
        """
        if trace is None:
            trace = self.default_trace
        if window is None:
            window = self.default_window
        if hasattr(trace, '__iter__'):
            return pd.concat(
                [self.fetch_trace(t, horizontal=horizontal) for t in trace]
            )

        if horizontal:
            index = self.fetch_horizontal(trace)
            values = self.query_ieee_array(f'TRAC{window}:DATA? TRACE{trace}')
            return pd.DataFrame(values, columns=['Trace ' + trace], index=index)
        else:
            values = self.query_ieee_array(f'TRAC{window}:DATA? TRACE{trace}')
            return pd.DataFrame(values)

    def fetch_timestamps(self, window=None, all=True, timeout=50000) -> NumpyArrayType:
        """Fetch data timestamps associated with acquired data. Not all types of acquired data support timestamping,
        and not all modes support the trace argument. A choice that is incompatible with the current state
        of the signal analyzer should lead to a TimeoutError.

        Arguments:
            all: If True, acquire and return all available timestamps; if False, only the most current timestamp.
            window: The window number corresponding to the desired timestamp data (or self.default_window when window=None)
        """

        if window is None:
            window = self.default_window

        if all:
            _to, self.backend.timeout = self.backend.timeout, timeout

        try:
            scope = 'ALL' if all else 'CURR'
            timestamps = self.backend.query_ascii_values(
                f'CALC{window}:SGR:TST:DATA? {scope}', container=np.array
            )
            timestamps = timestamps.reshape((timestamps.shape[0] // 4, 4))[:, :2]
            ret = timestamps[:, 0] + 1e-9 * timestamps[:, 1]
        finally:
            if all:
                self.backend.timeout = _to

        if not all:
            return ret[0]
        else:
            return ret

    def fetch_spectrogram(
        self,
        window: Union[int, None] = None,
        freqs: str = 'exact',
        timestamps: str = 'exact',
        timeout=None,
    ):
        """
        Fetch a spectrogram without initiating a new trigger. This has been tested in IQ Analyzer and real time
        spectrum analyzer modes. Not all instrument operating modes support trace selection; a choice that is
        incompatible with the current state of the signal analyzer should lead to a TimeoutError.

        Arguments:
            freqs: 'exact' (to fetch the frequency axis), 'fast' (to guess at index values based on FFT parameters), or None (leaving the integer indices)
            timestamps: 'exact' (to fetch the frequency axis), 'fast' (to guess at index values based on sweep time), or None (leaving the integer indices)
            window: The window number corresponding to the desired timestamp data (or self.default_window when window=None)
        :return: a pandas DataFrame containing the acquired data
        """
        if timeout is None:
            if self.trigger_source.lower() == 'mask':
                max_trigger_time = self.trigger_post_time
            else:
                max_trigger_time = self.sweep_dwell_time

            old_timeout, self.backend.timeout = (
                self.backend.timeout,
                6 * 1e3 * max_trigger_time,
            )
        else:
            old_timeout = timeout

        with self.suppress_timeout():
            if window is None:
                window = self.default_window

            data = self.query_ieee_array(f'TRAC{window}:DATA? SPEC')

            # Fetch time axis
            if timestamps not in ('fast', 'exact', None):
                raise ValueError("timestamps argument must be 'fast', 'exact', or None")
            elif timestamps == 'exact':
                t = self.fetch_timestamps(all=True, window=window, timeout=timeout)
            elif timestamps is None:
                t = None

            # Fetch frequency axis
            if freqs == 'fast':
                fc = self.frequency_center
                fsamp = self.iq_sample_rate
                Nfreqs = self.sweep_points
                f_ = fc + np.linspace(
                    -fsamp * (1.0 - 1.0 / Nfreqs) / 2,
                    +fsamp * (1.0 - 1.0 / Nfreqs) / 2,
                    Nfreqs,
                )
            if freqs == 'exact':
                f_ = self.fetch_horizontal(window)
                Nfreqs = len(f_)
            elif freqs is None:
                f_ = None
                Nfreqs = self.sweep_points

            # Reshape data according to frequency axis, since we'll be most certain
            # to know that dimension
            if data.size > 1:
                data = data.reshape((data.size // Nfreqs, Nfreqs))

            # Generate timestamps if we're going to guesstimate
            if timestamps == 'fast':
                if window == 1:
                    sweep_time = self.sweep_time
                else:
                    sweep_time = getattr(self, 'sweep_time_window' + window)
                ts0 = self.fetch_timestamps(all=False, window=window)
                t = (ts0 - sweep_time * data.shape[0]) + sweep_time * np.arange(
                    data.shape[0]
                )[::-1]

            self.backend.timeout = old_timeout

            if data.size > 1:
                return pd.DataFrame(
                    data[::-1], columns=f_, index=None if t is None else t[::-1]
                )
            else:
                return pd.DataFrame([], columns=f_)

        self.backend.timeout = old_timeout

    def fetch_marker(
        self, marker: int, axis: Union[Literal['X'], Literal['Y']]
    ) -> float:
        """Get marker value

        Arguments:
            marker: marker number on instrument display
            axis: 'X' for x axis or 'Y' for y axis
        """
        return float(self.query(f'CALC:MARK{marker}:{axis}?'))

    def get_marker_enables(self) -> DataFrameType:
        markers = list(range(1, 17))
        states = [
            [
                self.query(f'CALC:MARK{m}:STATE?'),
                self.query(f'CALC:MARK{m}:FUNC:BPOW:STATE?'),
            ]
            for m in markers
        ]

        df = pd.DataFrame(states, columns=['Marker', 'Band'], index=markers)

        df.index.name = 'Marker'

        return df.astype(int).astype(bool)

    def get_marker_power(self, marker: int) -> float:
        """Get marker value (on vertical axis)

        Arguments:
            marker: marker number on instrument display
        """
        return float(self.query(f'CALC:MARK{marker}:Y?'))

    def get_marker_position(self, marker: int) -> float:
        """Get marker position (on horizontal axis)

        Arguments:
            marker: marker number on instrument display

        Returns:
            position of the marker, in units of the horizontal axis
        """
        return float(self.query(f'CALC:MARK{marker}:X?'))

    def set_marker_position(self, marker: int, position: float):
        """get marker position (on horizontal axis)

        Arguments:
            marker: marker number on instrument display
            position: position of the marker, in units of the horizontal axis
        """
        return self.write(f'CALC:MARK{marker}:X {position}')

    def trigger_output_pulse(self, port: int):
        """trigger a pulse on a trigger output

        Arguments:
            port: Trigger port number

        :return: None
        """
        self.write(f'OUTPUT:TRIGGER{port}:PULS:IMM')


class _RSSpectrumAnalyzerMixIn(RohdeSchwarzFSWBase):
    expected_channel_type = attr.value.str('SAN', inherit=True)

    def get_marker_band_power(self, marker: int) -> float:
        """Get marker band power measurement

        Arguments:
            marker: marker number on instrument display

        Returns:
            power in dBm
        """

        return float(self.query(f'CALC:MARK{marker}:FUNC:BPOW:RES?'))

    def get_marker_band_span(self, marker: int) -> float:
        """Get span of marker band power measurement

        Arguments:
            marker: marker number on instrument display

        Returns:
            bandwidth in Hz
        """
        return float(self.query(f'CALC:MARK{marker}:FUNC:BPOW:SPAN?'))

    def get_marker_power_table(self):
        """Get the values of all markers."""
        enables = self.get_marker_enables()

        values = pd.DataFrame(
            columns=['Frequency'] + enables.columns.values.tolist(),
            index=enables.index,
            dtype=float,
            copy=True,
        )

        for m in enables.index:
            if enables.loc[m, 'Marker']:
                values.loc[m, 'Marker'] = self.get_marker_power(m)
                values.loc[m, 'Frequency'] = self.get_marker_position(m)

                if enables.loc[m, 'Band']:
                    values.loc[m, 'Band'] = self.get_marker_band_power(m)

        values.dropna(how='all', inplace=True)
        return values

    def fetch_marker_bpow(self, marker: int) -> float:
        """Get marker band power measurement

        Arguments:
            marker: marker number on instrument display

        """

        mark_cmd = 'CALC:MARK' + str(marker) + ':FUNC:BPOW:RES?'
        marker_val = float(self.query(mark_cmd))
        return marker_val

    def fetch_marker_bpow_span(self, marker: int) -> float:
        """Get marker band power measurement

        Arguments:
            marker: marker number on instrument display
        """

        return float(self.query(f'CALC:MARK{marker}:FUNC:BPOW:SPAN?'))


class _RSLTEAnalyzerMixIn(RohdeSchwarzFSWBase):
    format = attr.property.str(key='FORM', only=('REAL', 'ASCII'), case=False)

    @attr.property.float(min=0)
    def uplink_sample_rate(self):
        response = self.query('CONF:LTE:UL:BW?')
        return float(response[2:].replace('_', '.')) * 1e6

    @attr.property.float(min=0)
    def downlink_sample_rate(self):
        response = self.query('CONF:LTE:DL:BW?')
        return float(response[2:].replace('_', '.')) * 1e6

    def open(self):
        lb.VISADevice.open(self)
        # self.verify_channel_type()
        self.format = 'REAL'

    def fetch_power_vs_symbol_x_carrier(self, window, trace):
        data = self.fetch_trace(window=window, trace=trace)

        # Dimensioning is based on LTE standard definitions
        # of the resource block
        Ncarrier = int(50 * self.uplink_sample_rate / 10e6)
        Nsubcarrier = 12 * Ncarrier
        Nsymbol = data.size // Nsubcarrier

        data[data > 1e30] = np.nan

        data = data.values.reshape((Nsymbol, Nsubcarrier))
        data = pd.DataFrame(data)
        data.index.name = 'Symbol'
        data.columns.name = 'Subcarrier'
        return data

    # The methods should be deprecatable now, since the base fetch_trace() should work fine now that format is set
    # in the connect() method and the window parameter is supported by fetch_trace()
    def get_ascii_window_trace(self, window, trace):
        self.write('FORM ASCII')
        data = self.backend.query_ascii_values(
            f'TRAC{window}:DATA? TRACE{trace}', container=pd.Series
        )
        return data

    def get_binary_window_trace(self, window, trace):
        self.write('FORM REAL')
        data = self.backend.query_binary_values(
            f'TRAC{window}:DATA? TRACE{trace}',
            datatype='f',
            is_big_endian=False,
            container=pd.Series,
        )
        return data

    def get_allocation_summary(self, window):
        self.write('FORM ASCII')
        data = self.query(f'TRAC{window}:DATA? TRACE1').split(',')
        return data


class _RSIQAnalyzerMixIn(RohdeSchwarzFSWBase):
    _IQ_FORMATS = ('FREQ', 'MAGN', 'MTAB', 'PEAK', 'RIM', 'VECT')
    _IQ_MODES = ('TDOMain', 'FDOMain', 'IQ')

    expected_channel_type = attr.value.str('RTIM', inherit=True)

    iq_simple_enabled = attr.property.bool(key='CALC:IQ')
    iq_evaluation_enabled = attr.property.bool(key='CALC:IQ:EVAL')
    iq_mode = attr.property.str(key='CALC:IQ:MODE', only=_IQ_MODES, case=False)
    iq_record_length = attr.property.int(key='TRAC:IQ:RLEN', min=1, max=461373440)
    iq_sample_rate = attr.property.float(key='TRAC:IQ:SRAT', min=1e-9, max=160e6)
    iq_format = attr.property.str(key='CALC:FORM', only=_IQ_FORMATS, case=False)
    iq_format_window2 = attr.property.str(
        key='CALC2:FORM', case=False, only=_IQ_FORMATS
    )

    def fetch_trace(self, horizontal=False, trace=None):
        fmt = self.iq_format
        if fmt == 'VECT':
            df = RohdeSchwarzFSWBase.fetch_trace(self, horizontal=False, trace=trace)
        else:
            df = RohdeSchwarzFSWBase.fetch_trace(
                self, horizontal=horizontal, trace=trace
            )

        if fmt == 'RIM':
            if hasattr(df, 'columns'):
                df = pd.DataFrame(
                    df.iloc[: len(df) // 2].values
                    + 1j * df.iloc[len(df) // 2 :].values,
                    index=df.index[: len(df) // 2],
                    columns=df.columns,
                )
            else:
                df = pd.Series(
                    df.iloc[: len(df) // 2].values
                    + 1j * df.iloc[len(df) // 2 :].values,
                    index=df.index[: len(df) // 2],
                )
        if fmt == 'VECT':
            df = pd.DataFrame(df.iloc[1::2].values, index=df.iloc[::2].values)

        return df

    def store_trace(self, path):
        self.write(f"MMEM:STOR:IQ:STAT 1, '{path}'")


class _RSRealTimeMixIn(RohdeSchwarzFSWBase):
    TRIGGER_SOURCES = 'IMM', 'EXT', 'EXT2', 'EXT3', 'MASK', 'TDTR'
    WINDOW_FUNCTIONS = 'BLAC', 'FLAT', 'GAUS', 'HAMM', 'HANN', 'KAIS', 'RECT'
    _BOOL_LABELS = {False: '0', True: '1'}
    expected_channel_type = attr.value.str('RTIM', inherit=True)

    trigger_source = attr.property.str(
        key='TRIG:SOUR', only=TRIGGER_SOURCES, case=False
    )
    trigger_post_time = attr.property.float(key='TRIG:POST', min=0)
    trigger_pre_time = attr.property.float(key='TRIG:PRET', min=0)

    iq_fft_length = attr.property.int(key='IQ:FFT:LENG', sets=False)
    iq_bandwidth = attr.property.float(key='TRAC:IQ:BWID', sets=False)
    iq_sample_rate = attr.property.float(key='TRACe:IQ:SRAT', sets=False)
    iq_trigger_position = attr.property.float(key='TRAC:IQ:TPIS', sets=False)

    sweep_dwell_auto = attr.property.bool(key='SWE:DTIM:AUTO')
    sweep_dwell_time = attr.property.float(key='SWE:DTIM', min=30e-3)
    sweep_window_type = attr.property.str(
        key='SWE:FFT:WIND:TYP', case=False, only=WINDOW_FUNCTIONS
    )

    def store_spectrogram(self, path, window=2):
        self.mkdir(os.path.split(path)[0])
        self.write(f"MMEM:STOR{window}:SGR '{path}'")

    def clear_spectrogram(self, window=2):
        self.write(f'CALC{window}:SGR:CLE')

    def fetch_horizontal(self, window=2, trace=1):
        return self.backend.query_binary_values(
            f'TRAC{window}:X? TRACE{trace}', datatype='f', container=np.array
        )

    def set_detector_type(self, type_, window=None, trace=None):
        if window is None:
            window = self.default_window
        if trace is None:
            trace = self.default_trace
        self.write(f'WIND{window}:DET{trace} {type_}')

    def get_detector_type(self, window=None, trace=None):
        if window is None:
            window = self.default_window
        if trace is None:
            trace = self.default_trace
        return self.query('WIND{window}:DET{trace}?')

    def set_spectrogram_depth(self, depth, window=None):
        if window is None:
            window = self.default_window

        self.write(f'CALC{window}:SPEC:HDEP {depth}')

    def get_spectrogram_depth(self, window=None):
        if window is None:
            window = self.default_window

        return self.query(f'CALC{window}:SPEC:HDEP?')

    @attr.property.float(max=0)
    def trigger_mask_threshold(self):
        return self.get_frequency_mask(first_threshold_only=True)

    @trigger_mask_threshold.setter
    def _(self, thresholds):
        """'defined in dB relative to the reference level"""
        self.set_frequency_mask(thresholds, None)

    def set_frequency_mask(
        self,
        thresholds: list[float],
        frequency_offsets: list[float] = None,
        kind=Union[Literal['upper'], Literal['lower']],
        window: int = None,
    ):
        """Define the frequency-dependent trigger threshold values for a frequency mask trigger.

        Arguments:
            thresholds: trigger threshold at each frequency in db relative to the reference level (same size as `frequency_offsets`), or a scalar to use a constant value across the band
            array-like frequency_offsets: frequencies at which the mask is defined, or None (to specify across the whole band)
            kind: either 'upper' or 'lower,' corresponding to a trigger on entering the upper trigger definition or on leaving the lower trigger definition
            window: The window number corresponding to the desired trigger setting (or self.default_window when window=None)
        :return: None
        """
        if window is None:
            window = self.default_window
        if kind.lower() not in ('upper', 'lower'):
            raise ValueError(
                f'frequency mask is "{kind}" but must be "upper" or "lower"'
            )
        if frequency_offsets is None:
            bw = self.iq_bandwidth
            frequency_offsets = -bw / 2, bw / 2

            if not hasattr(thresholds, '__iter__'):
                thresholds = len(frequency_offsets) * [thresholds]
            elif len(thresholds) != 2:
                raise ValueError(
                    'with frequency_offsets=None, thresholds must be a scalar or have length 2'
                )

        self.write(f"CALC{window}:MASK:CDIR '.'")
        self.write(f"CALC{window}:MASK:NAME '{DEFAULT_CHANNEL_NAME}'")
        flat = np.array([frequency_offsets, thresholds]).T.flatten()

        plist = ','.join(flat.astype(str))

        self.write(f'CALC{window}:MASK:{kind} {plist}')

    def get_frequency_mask(
        self,
        kind: Union[Literal['upper'], Literal['lower']] = 'upper',
        window: int = None,
        first_threshold_only: bool = False,
    ):
        """Define the frequency-dependent trigger threshold values for a frequency mask trigger.

        Arguments:
            kind: either 'upper' or 'lower,' corresponding to a trigger on entering the upper trigger definition or on leaving the lower trigger definition
            window: The window number corresponding to the desired trigger setting (or self.default_window when window=None)
            bool first_threshold_only: if True, return only the threshold; otherwise, return a complete parameter dict
        :return: the threshold or a dictionary with keys "frequency_offsets" and "thresholds" and corresponding values (in Hz and dBm, respectively) of equal length
        """

        if window is None:
            window = self.default_window
        if kind.lower() not in ('upper', 'lower'):
            raise ValueError(
                f'frequency mask is "{kind}" but must be "upper" or "lower"'
            )

        plist = self.query(f'CALC{window}:MASK:{kind}?')
        plist = np.array(plist.split(',')).astype(np.float64)

        if first_threshold_only:
            return plist[1]
        else:
            return {'frequency_offsets': plist[::2], 'thresholds': plist[1::2]}

    def setup_spectrogram(
        self,
        center_frequency: float,
        analysis_bandwidth: float,
        reference_level: float,
        time_resolution: float,
        acquisition_time: float,
        input_attenuation: Union[float, None] = None,
        trigger_threshold: Union[float, None] = None,
        detector: Union[Literal['SAMP'], Literal['AVER']] = 'SAMP',
        analysis_window: Union[str, None] = None,
    ):
        """Quick setup for a spectrogram measurement in RTSA mode.

        Arguments:
            center_frequency: in Hz
            frequency_span: in Hz
            reference_level: in dBm
            time_resolution: in s
            acquisition_time: in s
            input_attenuation: in dB (or None to autoset based on the reference level)
            trigger_threshold: in dB (or None to free run)
            detector: 'SAMP' or 'AVER'
            analysis_window: one of ['BLAC','FLAT','GAUS','HAMM','HANN','KAIS','RECT']
        :return:
        """

        self.default_window = 2
        self.default_trace = 1

        if self.load_cache():
            return

        # Work around an apparent SCPI bug by setting
        # frequency parameters in spectrum analyzer mode
        with self.overlap_and_block(timeout=10000):
            self.apply_channel_type('SAN')
        self.channel_preset()
        self.wait()
        self.frequency_center = center_frequency
        self.wait()
        with self.overlap_and_block(timeout=10000):
            self.apply_channel_type('RTIM')
        self.wait()
        self.initiate_continuous = False
        self.frequency_span = analysis_bandwidth

        # Setup traces
        self.remove_window(1)

        # Level
        self.reference_level(reference_level, trace=1)
        if input_attenuation is not None:
            self.input_attenuation = input_attenuation
            if reference_level <= -30 and input_attenuation <= 5:
                self.input_preamplifier_enabled = True

        # Trace acquisition parameters
        self.set_detector_type(detector)

        # Sweep parameters
        self.sweep_time_window2 = time_resolution
        if self.sweep_time_window2 != time_resolution:
            self._logger.warning(
                f'requested time resolution {time_resolution}, '
                f'but instrument adjusted to {self.sweep_time_window2}'
            )
        self.spectrogram_depth = 100000

        # Triggering
        if trigger_threshold is not None:
            self.trigger_source = 'MASK'
            th = 2 * [trigger_threshold] + [0] + 2 * [trigger_threshold]
            rbw = 2 * self.resolution_bandwidth
            fr = [-analysis_bandwidth / 2, -rbw, 0, rbw, analysis_bandwidth / 2]
            self.set_frequency_mask(thresholds=th, frequency_offsets=fr)
            self.trigger_post_time = acquisition_time
            self.trigger_pre_time = 0.0
            self.trigger_post_time = acquisition_time
        else:
            self.sweep_dwell_auto = False
            self.sweep_dwell_time = acquisition_time

        # TODO: Parameterize somehow
        for i in (2, 3):
            self.output_trigger_direction('OUTP', output_trigger_index=i)
            self.output_trigger_type('DEV', output_trigger_index=i)
        if analysis_window is not None:
            self.sweep_window_type = 'BLAC'

        with self.overlap_and_block(timeout=2500):
            self.save_cache()
        lb.sleep(0.05)

        with self.overlap_and_block():
            self.spectrogram_depth

    def acquire_spectrogram_sequence(
        self, loop_time=None, delay_time=0.1, timestamps='fast'
    ):
        """Trigger and fetch data, optionally in a loop that continues for a specified
        duration.

        Arguments:
            loop_time: time (in s) to spend looping repeated trigger-fetch cycles, or None to execute once
            delay_time: delay time before starting (in s)
            timestamps: 'fast' (with potential for rounding errors to ~ 10 ns) or 'exact' (slow and not recommended)
        :return: dictionary structured as {'spectrogram': pd.DataFrame, 'spectrogram_acquisition_time': float, 'spectrogram_active_time': float}
        """
        specs = []

        if self.trigger_source.lower() == 'mask':
            max_trigger_time = self.trigger_post_time
        else:
            max_trigger_time = self.sweep_dwell_time
        t0 = time.time()
        time_remaining = loop_time
        active_time = 0

        while loop_time is None or time_remaining > 0:
            # Setup
            self.clear_spectrogram()
            self.wait()
            # Give the power sensor time to arm
            lb.sleep(delay_time)

            t0_active = time.time()
            # Try to trigger; block until timeout.
            with self.suppress_timeout():
                # print(max(30*max_trigger_time,time_remaining or 0.1))
                with self.overlap_and_block(timeout=int(1e3 * 30 * max_trigger_time)):
                    self.trigger_single(wait=False)
                lb.sleep(0.05)
                active_time += time.time() - t0_active

                self.backend.timeout = 6 * 1e3 * max_trigger_time
                single = self.fetch_spectrogram(
                    timeout=self.backend.timeout, timestamps=timestamps
                )
                self.backend.timeout = 1000
                if single is not None:
                    specs.append(single)
                self.wait()
            self.abort()

            if loop_time is None:
                break
            else:
                time_remaining = loop_time - (time.time() - t0)

        if len(specs) > 0:
            specs = pd.concat(specs, axis=0) if specs else pd.DataFrame().iloc[:, :-1]
        else:
            specs = pd.DataFrame()
            self._logger.warning('no data acquired')

        return {
            'spectrogram_data': specs,
            'spectrogram_acquisition_time': time.time() - t0,
            'spectrogram_active_time': active_time,
        }

    def arm_spectrogram(self):
        self.clear_spectrogram()
        self.wait()

    def acquire_spectrogram(self):
        """Trigger and acquire data, optionally in a loop that continues for a specified
        duration.

        :return: dictionary structured as {'spectrogram_active_time': float}
        """
        t0 = time.time()

        if self.trigger_source.lower() == 'mask':
            max_trigger_time = self.trigger_post_time
        else:
            max_trigger_time = self.sweep_dwell_time

        # Try to trigger; block until timeout.
        # print(max(30*max_trigger_time,time_remaining or 0.1))
        with self.overlap_and_block(timeout=int(1e3 * 30 * max_trigger_time)):
            self.trigger_single(wait=False)

        return {'spectrogram_active_time': time.time() - t0}


class RohdeSchwarzFSW26Base(RohdeSchwarzFSWBase):
    frequency_center = attr.property.float(max=26.5e9, inherit=True)
    frequency_span = attr.property.float(max=26.5e9, inherit=True)
    frequency_start = attr.property.float(max=26.5e9, inherit=True)
    frequency_stop = attr.property.float(max=26.5e9, inherit=True)


class RohdeSchwarzFSW26SpectrumAnalyzer(
    RohdeSchwarzFSW26Base, _RSSpectrumAnalyzerMixIn
):
    pass


class RohdeSchwarzFSW26LTEAnalyzer(RohdeSchwarzFSW26Base, _RSLTEAnalyzerMixIn):
    pass


class RohdeSchwarzFSW26IQAnalyzer(RohdeSchwarzFSW26Base, _RSIQAnalyzerMixIn):
    pass


class RohdeSchwarzFSW26RealTime(RohdeSchwarzFSW26Base, _RSRealTimeMixIn):
    pass


class RohdeSchwarzFSW43Base(RohdeSchwarzFSWBase):
    frequency_center = attr.property.float(max=43.5e9, inherit=True)
    frequency_span = attr.property.float(max=43.5e9, inherit=True)
    frequency_start = attr.property.float(max=43.5e9, inherit=True)
    frequency_stop = attr.property.float(max=43.5e9, inherit=True)


class RohdeSchwarzFSW43SpectrumAnalyzer(
    RohdeSchwarzFSW43Base, _RSSpectrumAnalyzerMixIn
):
    pass


class RohdeSchwarzFSW43LTEAnalyzer(RohdeSchwarzFSW43Base, _RSLTEAnalyzerMixIn):
    pass


class RohdeSchwarzFSW43IQAnalyzer(RohdeSchwarzFSW43Base, _RSIQAnalyzerMixIn):
    pass


class RohdeSchwarzFSW43RealTime(RohdeSchwarzFSW43Base, _RSRealTimeMixIn):
    pass



""" Driver classes for signal generators.
:author: Ryan Jacobs <ryan.jacobs@nist.gov>, Aziz Kord <aziz.kord@nist.gov>, Daniel Kuester <daniel.kuester@nist.gov>
Paul.Blanchard <paul.blanchard@nist.gov>
"""

import labbench as lb
from labbench import paramattr as attr

__all__ = ['RohdeSchwarzSMW200A']


@attr.visa_keying(remap={True: '1', False: '0'})
class RohdeSchwarzSMW200A(lb.VISADevice):
    make=attr.value.str('Rohde&Schwarz', inherit=True)
    model=attr.value.str('SMW200A', inherit=True)

    frequency_center = attr.property.float(
        key=':freq', min=2e3, max=26.5e9, step=1e3, label='Hz'
    )
    rf_output_power = attr.property.float(
        key=':pow', min=-145, max=20, step=1e-2, label='dBm'
    )
    rf_output_enable = attr.property.bool(key='OUTP')
    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    def save_state(self, FileName, num='4'):
        """Save current state of the device to the default directory.
        :param FileName: state file location on the instrument
        :type FileName: string

        :param num: state number in the saved filename
        :type num: int

        """
        self.write('MMEMory:STORe:STATe {},"{}.savrcltxt"'.format(num, FileName))

    def load_state(self, FileName, opc=False, num='4'):
        """Loads a previously saved state file in the instrument

        :param FileName: state file location on the instrument
        :type FileName: string

        :param opc: set the VISA op complete flag?
        :type opc: bool

        :param num: state number in the saved filename
        :type num: int
        """
        cmd = "MMEM:LOAD:STAT {},'{}.savrcltxt';*RCL {}".format(num, FileName, num)
        self.write(cmd, opc=opc)


# Example code works nicely in this if statement, which only runs if we're
# running this file (not if it's being imported by another file :))
if __name__ == '__main__':
    with RohdeSchwarzSMW200A('USB::2733::146::102240::0::INSTR') as siggen:
        siggen.rf_output_enable = True



__all__ = ['RigolTechnologiesMSO4014', 'TektronixMSO64B', 'TektronixMSO64BSpectrogram']

import labbench as lb
from labbench import paramattr as attr

scope_channel_kwarg = attr.method_kwarg.int(
    'channel', min=1, max=4, help='hardware input port'
)


@scope_channel_kwarg
class RigolTechnologiesMSO4014(lb.VISADevice):
    # used for automatic connection
    make = attr.value.str('RIGOL TECHNOLOGIES', inherit=True)
    model = attr.value.str('MSO4014', inherit=True)

    time_offset = attr.property.float(
        key=':TIM:OFFS',
        label='s',
        help='acquisition time offset relative to the trigger',
    )
    time_scale = attr.property.float(
        key=':TIM:SCAL', label='s', help='acquisition time per division'
    )
    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    def open(self, horizontal=False):
        self.write(':WAVeform:FORMat ASCii')

    def fetch(self):
        return self.backend.query_ascii_values(':WAV:DATA?')

    def fetch_rms(self):
        return float(self.backend.query(':MEAS:VRMS?').rstrip().lstrip())


@attr.visa_keying(remap={True: '1', False: '0'})
@scope_channel_kwarg
class TektronixMSO64B(lb.VISADevice):
    # used for automatic connection
    make = attr.value.str('TEKTRONIX', inherit=True)
    model = attr.value.str('MSO64B', inherit=True)
    timeout: float = attr.value.float(3)

    # horizontal acquisition
    record_length = attr.property.int(
        key='HORIZONTAL:RECORDLENGTH', label='samples', help='acquisition record length'
    )
    sample_rate = attr.property.float(key='HOR:SAMPLER', label='S/s', get_on_set=True)
    _horizontal_mode = attr.property.str(
        key='HORizontal:MODe',
        case=False,
        only=('AUTOMATIC', 'MANUAL'),
        help='AUTO automatically adjusts the sample rate and record length',
    )

    # input port parameters
    # TODO: check lower bound
    bandwidth = attr.method.float(
        key='CH{channel}:BANDWIDTH',
        min=0,
        max=8e9,
        label='Hz',
        help='input port analog bandwidth',
    )
    channel_enabled = attr.method.bool(
        key='SELect:CH{channel}',
        help='enable channel acquisition',
    )
    coupling = attr.method.str(
        key='CH{channel}:COUPLING',
        case=False,
        only=('AC', 'DC', 'DCReject'),
        cache=True,
        help='input port coupling mode',
    )
    input_termination = attr.method.float(
        key='CH{channel}:TERMINATION',
        only=(50, 1e6),
        cache=True,
        help='input port termination impedance',
    )
    # Trigger settings
    trig_type = attr.property.str(
        key='TRIGger:A:TYPE',  # Trigger A is hard coded, trigger B only applies to TRANSISTION trigger,\
        # indicated by 'Sequence' on the display
        only=(
            'EDGE',
            'WIDTH',
            'TIMEOUT',
            'RUNT',
            'WINDOW',
            'LOGIC',
            'SETHOLD',
        ),
        help='Trigger mode for captures',
    )
    # Trig width only associated with width type trigger
    trig_width_high = attr.property.float(
        key='TRIGger:A:PULSEWidth:HIGHLimit', min=4e-9, max=5
    )
    trig_width_low = attr.property.float(
        key='TRIGger:A:PULSEWidth:LOWLimit', min=8e-9, max=5
    )


    trig_edge_slope = attr.property.str(
        key='TRIGger:A:EDGE:SLOpe',
        only=("RISE", "FALL", "EITHER")
    )
    trig_edge_source = attr.property.str(
        key='TRIGger:A:EDGE:SOUrce',
        only=('CH1', 'CH2', 'CH3', 'CH4', 'LINE', 'AUX'),
        help='source for edge type trigger'
    )

    trig_level= attr.method.float(
        key='TRIGger:A:LEVel:CH{channel}',
    )

    trig_mode = attr.property.str(
        key='TRIGger:A:MODe',
        only=("AUTO", "NORMAL")
    )

    trig_holdoff_by = attr.property.str(
        key="TRIGger:A:HOLDoff:BY",
        only=("TIME", "RANDOM")
        )

    trig_holdoff_time = attr.property.float(
        key="TRIGger:A:HOLDoff:TIMe",
        min=0,
        max=10
    )

    # Acquisition settings
    acq_type = attr.property.str(
        key="ACQuire:STOPAfter",
        only=("RUNSTOP", "SEQUENCE")
    )
    acq_mode = attr.property.str(
        key="ACQuire:MODe",
        only=("SAMPLE", "PEAKDETECT", "HIRES", "AVERAGE", "ENVELOPE")
    )
    acq_num_seq = attr.property.int(
        key="ACQuire:SEQuence:NUMSEQuence",
    )
    acq_time_source = attr.property.str(
        key="ROSc:SOUrce",
        only=("INTERNAL", "EXTERNAL", "TRACKING")
    )
    acq_fastframe = attr.property.str(
        key="HORizontal:FASTframe:STATE",
        only=("ON", "OFF")
    )

    # vertical scale
    vertical_scale = attr.method.float(
        key='CH{channel}:SCALE',
        label='V',
        help='vertical scale of the specified channel',
    )

    # Data storage, specifically for waveform retrieval over VISA (not implemented)
    data_source = attr.property.str(
        key='DATa:SOUrce',
        only=('CH1',
              'CH2',
              'CH3',
              'CH4'),
        help='data source for acquisition'
    )
    data_start = attr.property.int(
        key='DATa:START',
        help='data collection start point'
    )
    data_stop = attr.property.int(
        key='DATa:STOP',
        help='data collection stop point'
    )

    # File operations
    file_copy = attr.property.str(
        key='FILESystem:COPy', help='copy a file within filesystem'
    )
    file_cwd = attr.property.str(
        key='FILESystem:CWD',
        help='current working directory'
        )
    file_read = attr.property.str(
        key='FILESystem:READFile',
        help='read the file over interface',
    )

    # Store/load
    def load_setup(self, setup_file_name: str):
        self.write(f"RECAll:SETUp \"{setup_file_name}\"")

    def retrieve_waveform(self, channel, tekhsi_port=5000):
        """
        Retrieves the next waveform base on the scope's current acquisition settings.

        args:
        channel -- specifiy the channel number 1-4
        tekhsi_port -- typically 5000, unless it is changed manually on the scope

        returns:
        wfm -- AnalogWaveform type with data, further manipulation needed. See tekhsi docs.
        """
        try:
            from tekhsi import TekHSIConnect, AcqWaitOn
            from tm_data_types import AnalogWaveform
        except ImportError as e:
            raise e("When using the TektronixMSO64B class, ensure tekhsi library is installed.")

        ip_addr = self.resource.split("::")[1]
        self._logger.info(f"Waiting for data on {ip_addr}:{tekhsi_port}")
        with TekHSIConnect(f"{ip_addr}:{tekhsi_port}", [f"ch{channel}"]) as connect:
            with connect.access_data(AcqWaitOn.AnyAcq):
                wfm: AnalogWaveform = connect.get_data(f"ch{channel}")
        return wfm

    def open(self):
        self._horizontal_mode = 'MANUAL'
        self.write(':VERBose 0')
        self.write(':HEADer 0')


class TektronixMSO64BSpectrogram(TektronixMSO64B):
    resolution_bandwidth = attr.property.float(
        key='SV:RBW', label='Hz', help='resolution bandwidth of all channels'
    )
    span = attr.property.float(
        key='SV:SPAN', max=2e9, label='Hz', help='analysis bandwidth per channel'
    )

    # don't log instrument front-end display settings that don't impact the data (log=False)
    _power_scale_min = attr.property.float(
        key='SV:SPECtrogram:CSCale:MIN',
        min=-170,
        max=99,
        log=False,
        label='dBm',
        help='spectrogram display power scale minimum',
    )
    _power_scale_max = attr.property.float(
        key='SV:SPECtrogram:CSCale:MAX',
        min=-170,
        max=99,
        log=False,
        label='dBm',
        help='spectrogram display power scale maximum',
    )

    # assume that these are only set once, so relegate them to metadata (cache=True)
    _sync_center_frequencies = attr.property.bool(
        key='SV:LOCKCenter',
        cache=True,
        help='whether to lock all channels to the same center frequency',
    )
    _resolution_bandwidth_mode = attr.property.str(
        key='SV:RBWMode',
        case=False,
        only=('AUTOMATIC', 'MANUAL'),
        cache=True,
        help='whether to automatically set analysis parameters to achieve a specified RBW',
    )

    center_frequency = attr.method.float(
        key='CH{channel}:SV:CENTERFrequency',
        max=8e9,
        label='Hz',
        help='channel center frequency',
    )

    bandwidth = attr.method.float(
        # TODO: check min
        key='CH{channel}:BANDWIDTH',
        min=0,
        max=8e9,
        label='Hz',
        help='channel analysis bandwidth',
    )

    # log=False here since we intend usage of spectrogram_enabled
    _only_spectrogram_enabled = attr.method.bool(
        key='SV:CH{channel}:SEL:SPEC',
        log=False,
        help='enable channel spectrogram analysis',
    )

    _only_spectrumview_enabled = attr.method.bool(
        key='CH{channel}:SV:STATE',
        log=False,
        help='enable channel spectrum view mode',
    )

    @attr.method.bool(help='channel center frequency')
    @scope_channel_kwarg
    def spectrogram_enabled(self, enabled: bool = lb.Undefined, /, *, channel: int):
        return all(
            [
                self.channel_enabled(enabled, channel),
                self._only_spectrumview_enabled(enabled, channel),
                self._only_spectrogram_enabled(enabled, channel),
            ]
        )

    def open(self):
        self._resolution_bandwidth_mode = 'MANUAL'
        self._sync_center_frequencies = False


if __name__ == '__main__':
    with TektronixMSO64B() as scope:
        # Load some setup file onboard the scope
        scope.load_setup("default_setup.set")
        # Manually set some params
        scope.sample_rate = 200e6
        scope.record_length = 20e6
        scope.trig_type = "EDGE"
        scope.trig_edge_source = "CH1"
        scope.trig_mode = "AUTO"
        scope.trig_holdoff_by = "RANDOM"
        scope.acq_type = "RUNSTOP"

        # Collect data from scope after an acquisition
        data = scope.retrieve_waveform(1)
        print(data.normalized_vertical_values)

    with TektronixMSO64BSpectrogram() as scope:
        scope.spectrogram_enabled(channel=1)

        scope.resolution_bandwidth = 10e3
        scope.span = 2e9

        scope.coupling('AC', channel=1)
        scope.bandwidth(8e9, channel=1)
        scope.input_termination(50, channel=1)
        scope.center_frequency(3e9, channel=1)
        scope.vertical_scale(0.05, channel=1)



import labbench as lb
from labbench import paramattr as attr

__all__ = ['RigolDP800Series']


class dp800_property_adapter(attr.visa_keying):
    def get(self, device: lb.Device, scpi_key: str, trait_name=None):
        """This instrument expects keys to have syntax ":COMMAND? PARAM",
        instead of ":COMMAND PARAM?" as implemented in lb.VISADevice.

        Insert the "?" in the appropriate place here.
        """
        if ' ' in scpi_key:
            key = scpi_key.replace(' ', '? ', 1)
        else:
            key = scpi_key + '?'
        return device.query(key)

    def set(self, device: lb.Device, scpi_key: str, value, trait_name=None):
        """This instrument expects sets to have syntax :COMMAND PARAM,VALUE
        instead of :COMMAND attr.VALUE. as implemented in lb.VISADevice.

        Implement this behavior here.
        """
        if ' ' in scpi_key:
            key = f'{scpi_key},{value}'
        else:
            key = f'{scpi_key} {value}'
        return device.write(key.rstrip())


@dp800_property_adapter(remap={True: 'ON', False: 'OFF'})
class RigolDP800Series(lb.VISADevice):
    # properties accept the "key" argument and/or decorators for custom implementation
    enable1 = attr.property.bool(key=':OUTP CH1', help='enable channel 1 output')
    enable2 = attr.property.bool(key=':OUTP CH2', help='enable channel 2 output')
    enable3 = attr.property.bool(key=':OUTP CH3', help='enable channel 3 output')

    voltage_setting1 = attr.property.float(
        key=':SOUR1:VOLT', help='output voltage setting on channel 1'
    )
    voltage_setting2 = attr.property.float(
        key=':SOUR2:VOLT', help='output voltage setting on channel 2'
    )
    voltage_setting3 = attr.property.float(
        key=':SOUR3:VOLT', help='output voltage setting on channel 3'
    )

    voltage1 = attr.property.float(
        key=':MEAS:VOLT CH1', sets=False, help='output voltage reading on channel 1'
    )
    voltage2 = attr.property.float(
        key=':MEAS:VOLT CH2', sets=False, help='output voltage reading channel 2'
    )
    voltage3 = attr.property.float(
        key=':MEAS:VOLT CH3', sets=False, help='output voltage reading channel 3'
    )

    current1 = attr.property.float(
        key=':MEAS:CURR CH1', sets=False, help='current draw reading on channel 1'
    )
    current2 = attr.property.float(
        key=':MEAS:CURR CH2', sets=False, help='current draw reading on channel 2'
    )
    current3 = attr.property.float(
        key=':MEAS:CURR CH3', sets=False, help='current draw reading on channel 3'
    )

    @lb.retry(BaseException, 3)
    def open(self):
        """Poll *IDN until the instrument responds.
        Sometimes it needs an extra poke before it responds.
        """
        try:
            timeout, self.backend.timeout = self.backend.timeout, 0.2
            self.identity
        finally:
            self.backend.timeout = timeout


if __name__ == '__main__':
    import time

    lb.show_messages('debug')

    inst = RigolDP800Series('USB0::0x1AB1::0x0E11::DP8C180200079::INSTR')

    with inst:
        print(inst.identity)
        inst.enable1
        inst.voltage_setting1 = 15.0
        inst.enable1 = True
        time.sleep(0.1)
        print(inst.voltage1, inst.current1)



import labbench as lb
from labbench import paramattr as attr

__all__ = ['RohdeSchwarzZMBSeries']


class RohdeSchwarzZMBSeries(lb.VISADevice):
    """A network analyzer.

    Author: Audrey Puls
    """

    initiate_continuous = attr.property.bool(key='INITiate1:CONTinuous:ALL', help='')

    options = attr.property.str(
        key='*OPT', sets=False, cache=True, help='installed license options'
    )

    def clear(self):
        self.write('*CLS')

    def save_trace_to_csv(self, path, trace=1):
        """Save the specified trace to a csv file on the instrument.
        Block until the operation is finished.
        """
        # This with block causes the function not to return until
        # the instrument is done saving data
        with self.overlap_and_block(timeout=3):
            self.write(f"MMEM:STOR:TRAC:CHAN {trace}, '{path}'")

    def trigger(self):
        """Initiate a software trigger.

        Consider setting `state.initiate_continuous = False` first so that the
        instrument waits for this trigger before starting a sweep.
        """
        self.write(':INITiate:IMMediate')


if __name__ == '__main__':
    from ssmdevices.instruments import MiniCircuitsRCDAT
    import time
    import numpy as np

    # Example: Calibration sweep of an attenuator
    #############################SETUP CONNECTIONS#####################
    atten = MiniCircuitsRCDAT('11604210008')  # Set the attenuator serial number here
    na = RohdeSchwarzZMBSeries('TCPIP0::132.163.202.153::inst0::INSTR')

    lb.show_messages('debug')

    with na, atten:
        na.clear()
        na.initiate_continuous = False

        ####################SWEEP THROUGH ATTENUATION AND COLLECT DATA#################
        # Since we want to collect cal data on (uncalibrated) attenuator settings,
        # we work with atten.attenuation_setting instead of
        # atten.attenuation (which tries to apply calibration data)
        for atten.attenuation_setting in np.linspace(0, 110, num=441):
            # The name of the run, based on the attenuation setting
            name = str(atten.attenuation_setting).replace('.', 'pt')

            na.trigger()
            time.sleep(20)  # Pauses python to let the VNA finish a full sweep
            na.save_trace_to_csv(f'VA_{atten.resource}_{name}.csv')

        # All done!
        print(
            r"""ATTENTION USER: This program has completed. Your data is stored
                  on the VNA under: C:\Users\Public\Documents\Rhode-Schwarz\VNA
                  This data can be transfered to another computer via USB.
               """
        )



""" Network test instruments

Authors:
    Aziz Kord
    Dan Kuester
"""

import numbers
import os
import re
import time

import labbench as lb
from labbench import paramattr as attr

__all__ = ['AeroflexTM500']


class TM500Error(ValueError):
    def __init__(self, msg, errcode=None):
        super(TM500Error, self).__init__(msg)
        self.errcode = errcode


class AeroflexTM500(lb.TelnetDevice):
    """Control an Aeroflex TM500 network tester with a
    telnet connection.

    The approach here is to iterate through lines of bytes, and
    add delays as needed for special cases as defined in the
    `delays` attribute.

    At some point, these lines should just be loaded directly
    from a file that could be treated as a config file.
    """

    timeout: float = attr.value.float(1)
    ack_timeout: float = attr.value.float(
        default=30,
        min=0.1,
        help='how long to wait for a command acknowledgment from the TM500 (s)',
    )
    busy_retries: int = attr.value.int(default=20, min=0)
    remote_ip: str = attr.value.str(
        default='10.133.0.203', help='ip address of TM500 backend'
    )
    remote_ports: str = attr.value.str(
        default='5001 5002 5003', help='port of TM500 backend'
    )
    min_acquisition_time: int = attr.value.int(
        default=30, min=0, help='minimum time to spend acquiring logs (s)'
    )
    port: int = attr.value.int(default=5003, min=1)
    config_root: str = attr.value.str(
        default='.', help='path to the command scripts directory'
    )
    data_root: str = attr.value.str(default='.', help='remote save root directory')
    convert_files = attr.value.list(
        default=[], help='text to match in the filename of data output files to convert'
    )

    def arm(self, scenario_name):
        """Load the scenario from the command listing in a local TM500
        configuration file.
        The the full path to the configuration file is
        `os.path.join(self.config_root, self.config_file)+'.conf'`
        (on the host computer running this python instance).

        If the last script that was run is the same as the selected config
        script, then the script is loaded and sent to the TM500 only
        if force=True. It always runs on the first call after AeroflexTM500
        is instantiated.

        Returns:
            A list of responses to each command sent
        """
        if isinstance(scenario_name, numbers.Number):
            scenario_name = str(scenario_name)
        if scenario_name == self.__latest.setdefault('scenario_name', None) is not None:
            raise TM500Error(
                'the TM500 is already armed with the scenario named {}'.format(
                    scenario_name
                )
            )

        config_path = os.path.join(self.config_root, scenario_name) + '.conf'
        self._logger.debug('arming TM500 scenario {}'.format(repr(config_path)))

        t0 = time.time()
        with open(config_path, 'r') as f:
            seq = f.readlines()
        self._reconnect()
        ret = self._send(seq)
        if self.data_root is not None:
            self._send('#$$DATA_LOG_FOLDER 1 "{}"'.format(self.data_root))
        self._logger.debug('armed in {:.2f}s'.format(time.time() - t0))
        self.__latest['scenario_name'] = scenario_name
        return ret

    def trigger(self):
        """Start logging and return the path to the directory where the data
        is being saved.
        """
        if self.__latest.get('scenario_name') is None:
            raise TM500Error('arm before triggering')
        self._reconnect()
        self.__latest['data'] = self._send('#$$START_LOGGING').split("'")[1]
        self.__latest['trigger_time'] = time.time()
        return self.__latest['data']

    def stop(self, convert=True):
        """Stop logging.
        :param bool convert: Whether to convert the output binary files to text

        :returns: If convert=True, a dictionary of {'name': path} items pointing to the converted text output
        """
        ret = {} if convert else None
        self._block_until_min_acquisition()
        try:
            # Stop running
            self._reconnect()
            self._send('forw mte MtsClearMts')
            self._send('WAIT FOR "I: CMPI MTE 0 MTSCLEARMTS COMPLETE IND" TIMEOUT 60')
            self._send('forw mte DeConfigRdaStopTestCase')
            self._send('WAIT FOR "I: CMPI DTE RDA TEST GROUP STOPPED IND" TIMEOUT 300')
        except TM500Error as e:
            self._logger.debug(
                'exception on attempt to stop scenario: {}'.format(str(e))
            )
        else:
            if self.__latest.setdefault('data', None) is None:
                self._logger.debug('there is no active logging to stop')
            else:
                self._send('#$$STOP_LOGGING')
                if convert:
                    ret = self._convert_to_text()
                self._send('forw mte GetRrcStats [] [] [] [1] [1]')
                self._send('forw mte GetNasStats [] [] [] [] [1] [1]')
        try:
            self._send(
                'SDLI LTE_RADIO_CONTEXT 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000'
            )
            self._send(
                'SDLI LTE_NAS_STATUS 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000'
            )
            self._send(
                'SDLI LTE_RRC_STATUS 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000 0x00000000'
            )
        except TM500Error as e:
            self._logger.debug(
                'exception on attempt to cleanup scenario: {}'.format(str(e))
            )

        self.__latest['scenario_name'] = None
        self.__latest['data'] = None
        self.__latest['trigger_time'] = 0
        return ret

    def reboot(self, timeout=180):
        """Reboot the TMA and TM500 hardware."""
        self._send('RBOT')
        lb.TelnetDevice.close(self)
        t0 = time.time()
        while time.time() - t0 < timeout:
            try:
                lb.TelnetDevice.open(self)
                break
            except TimeoutError:
                lb.sleep(1)
                continue
            except:
                raise
        else:
            raise TimeoutError('timeout reconnecting to the tm500 after reboot')

        ret = b''
        while time.time() - t0 < timeout:
            ret = ret + self.backend.read_very_eager()

            if b'I: REBOOT - The Test Mobile has rebooted' in ret:
                break
            else:
                lb.sleep(1)
        else:
            raise TimeoutError('no message confirming reboot')

        self._logger.debug('tm500 rebooted after {}s'.format(time.time() - t0))

    @staticmethod
    def command_log_to_script(path):
        """Scrape a script out of a TM500 "screen save" text file. The output
        for an input that takes the form <path>/<to>/<filename>.txt
        will be <path>/<to>/<filename>-script.txt.
        """
        with open(path, 'rb') as f:
            txt = f.read()

        # Remove the timestamp.
        # Assumes tab-delimiter separtaes timestamps from the telnet traffic.
        txt = b'\r\n'.join([line.split(b'\t', 1)[-1] for line in txt.splitlines()])

        # start where things get interesting
        #        i = txt.lower().find(b'abot 0 0 0')
        lines = txt.splitlines()
        for i, line in enumerate(lines):
            if line.lower().strip().startswith(b'rset'):
                break
        else:
            raise ValueError('no start delimited by RSET command found in command log')
        txt = b'\r\n'.join(lines[i:])

        # end at activate
        i = txt.upper().find(b'C: FORW 0X00 OK MTE ACTIVATE')
        if i == -1:
            raise ValueError('no activate found in command log')
        i_eol = txt[i:].find(b'\r\n')
        if i_eol != -1:
            i = i + i_eol
        txt = txt[:i]

        blocks = re.split(b'(^C: .*)', txt, flags=re.MULTILINE | re.IGNORECASE)

        # Throw out odd-numbered leftover at the end
        blocks = blocks[: 2 * (len(blocks) // 2)]

        # The result is organized by pairs of (big chunk of data, command response).
        # Get the command from the first word of the incoming command
        commands = []
        for i in range(0, len(blocks), 2):
            # Get the 2nd field in the response field. That's the start of command
            # the full line command.
            rsp = blocks[i + 1].split()
            if len(rsp) > 1:
                cmd = rsp[1]
            else:
                print(
                    "\n\nWARNING: Don't know what to do with command response ",
                    blocks[i + 1],
                )

            # Find the line that starts with the line in the previous block of
            # text. Throw a warning unless there was exactly 1 match in the
            # block of text.
            matches = re.findall(
                rb'^([\#\$]*' + cmd + b'.*)', blocks[i], re.MULTILINE | re.IGNORECASE
            )
            if len(matches) == 0:
                print(
                    '\n\nWARNING: Found no command for ',
                    blocks[i + 1].rstrip().strip(),
                    '\n after ',
                    repr(blocks[i]),
                )
            elif len(matches) > 1:
                print(
                    '\n\nWARNING: More than one match for ',
                    blocks[i + 1].rstrip().strip(),
                    ': ',
                    repr(matches),
                )
            elif b'START_LOGGING' in matches[0].rstrip().upper():
                pass
            else:
                commands.append(matches[0].rstrip())

        with open(os.path.splitext(path)[0] + '.conf', 'wb') as f:
            f.write(b'\r\n'.join(commands))

    def close(self):
        try:
            if self.__latest.get('scenario_name') is not None:
                self.stop(convert=False)
        except TM500Error as e:
            if e.errcode not in (0x02, 0x06):
                raise
        finally:
            try:
                self._send('#$$DISCONNECT', confirm=False)
            except BaseException:
                pass

    def open(self):
        self.__latest = {}

        # Invalidate any incomplete previous commands in the remote telnet buffer
        try:
            self._send('***', timeout=1)
        except (ValueError, TimeoutError):
            pass

        try:
            self._send('#$$DISCONNECT', timeout=1)
        except BaseException:
            pass

        self._reconnect(force=True)

        # Set up licensing
        self._send('ABOT 0 0 0')
        self._send('SCFG SWL')
        self._send('STRT')
        self._send('forw swl setlicenseserver none')

    def _tma_is_connected(self):
        try:
            self._send('CHOW')
        except TM500Error as e:
            if e.errcode in (0x02, 0x06):
                return False
            else:
                raise
        return True

    def _reconnect(self, force=False):
        """Ensure the TMA software is connected to the TM500"""
        if not force and self._tma_is_connected():
            return
        else:
            # Now connect
            self._send(
                '#$$PORT {ip} {ports}'.format(
                    ip=self.remote_ip, ports=self.remote_ports
                ),
                timeout=1,
            )
            self._send('#$$CONNECT')

    def _block_until_min_acquisition(self):
        """Make sure the minimum acquisition time has elapsed to avoid putting
        the TM500 in a sad state
        """

        if self.__latest.get('trigger_time', 0) > 0:
            elapsed = time.time() - self.__latest.get('trigger_time', 0)
            if elapsed < self.min_acquisition_time:
                lb.sleep(self.min_acquisition_time - elapsed)

    def _send(self, msg, data_lines=1, confirm=True, timeout=None):
        """Send a message, then block until a confirmation message is received.

        :param msg: str or bytes containing the message to send
        :param int data_lines: number of lines in the response string

        :returns: decoded string containing the response
        """

        # First, if this is a sequence of messages, work through them one by one
        if hasattr(msg, '__iter__') and not isinstance(msg, (str, bytes)):
            return [self._send(m) for m in msg]

        # Send the message
        if isinstance(msg, str):
            msg = msg.encode('ascii')
        msg = msg.strip().rstrip()
        if len(msg) == 0:
            return ''
        elif msg.strip().lower().startswith(b'wait for') and b'timeout' in msg.lower():
            timeout = int(msg.lower().rsplit(b'timeout', 1)[1].strip())

        self._logger.debug('write {}'.format(repr(msg)))
        self.backend.write(msg + b'\r')

        if not confirm:
            return

        # Figure out the expected response
        rsp = msg.split(b' ', 1)[0]
        if rsp.startswith(b'#$$'):
            rsp = rsp[3:]
        rsp = b'C: ' + rsp + b' '

        # Block until the response
        try:
            ret = self._read_until(rsp, timeout)

            # Receive through the end of line for the rest of the status response
            if timeout is None:
                timeout = self.timeout
            ret = rsp
            for i in range(data_lines):
                ret = ret + self.backend.read_until(b'\r', timeout=timeout)
                code = int(ret.strip().split(b'0x', 1)[1].split(maxsplit=1)[0], 16)
                if i == 0 and code != 0:
                    raise TM500Error(
                        'Error in message {}: {}'.format(repr(msg), repr(ret)), code
                    )
        except TimeoutError:
            raise TimeoutError(
                'timeout waiting for response to command {}'.format(repr(msg))
            )

        # Receive any other data received during the delay
        extra = self.backend.read_very_eager().strip().rstrip()
        if extra:
            if extra.count(b'\r') <= 1:
                self._logger.debug('extra -> {}'.format(extra.decode()))
            else:
                self._logger.debug('extra -> ({} lines)'.format(extra.count(b'\r')))
        return ret.decode('ascii')

    def _convert_to_text(self):
        """Convert the latest data to text"""
        root = self.__latest['data']

        t0 = time.time()
        lb.sleep(0.5)  # TODO Is this really necessary?

        #            pre_entries = [f for f in os.listdir(root)]
        self._send('#$$CONVERT_TO_TEXT', timeout=30)
        entries = [f for f in os.listdir(root) if f.lower().endswith(('csv', 'txt'))]
        names = ['_'.join(e.split('_')[3:-1]).replace('-', '_') for e in entries]
        paths = [os.path.join(root, e) for e in entries]
        ret = dict(zip(names, paths))

        # Remove items not listed in self.convert_files
        if len(self.convert_files) > 0:
            for name in list(ret.keys()):
                for inc in self.convert_files:
                    if inc.upper() in name.upper():
                        break
                else:
                    del ret[name]

        lb.sleep(0.5)

        self._logger.info(
            'converted TM500 logs from binary in {:0.2f}s'.format(time.time() - t0)
        )

        return ret

    def _read_until(self, rsp, timeout=None):
        if timeout is None:
            ack_timeout = self.ack_timeout
        else:
            ack_timeout = timeout
        #        self._logger.debug('awaiting response {}'.format(repr(rsp)))

        # Block until the expected response is received
        t0 = time.time()
        rsp = rsp.upper()
        ret = b''
        while time.time() - t0 < ack_timeout:
            # Looping on short blocking makes it easier to
            # cancel execution with a KeyboardInterrupt
            ret += self.backend.read_until(rsp, ack_timeout)
            if rsp in ret:
                break
        else:
            raise TimeoutError('response timeout')
        #        self._logger.debug('    -> {}'.format(ret))
        return ret


if __name__ == '__main__':
    import labbench as lb

    AeroflexTM500.key_log_to_script(r'C:\Users\dkuester\Desktop\TM500_2Sec_8UEs_withTime.txt')

    path = r'e:\TM500ScriptForPaulDebug'
    lb.show_messages('debug')
    tm500 = AeroflexTM500('10.133.0.202')
    with tm500, lb.stopwatch():
        tm500.configure(path)


"""
Made by Michael Voecks
"""

import labbench as lb
from labbench import paramattr as attr
import ssmdevices.lib


class AndroidDebugBridge(lb.ShellBackend):
    binary_path = attr.value.Path(ssmdevices.lib.path('adb.exe'), inherit=True)
    timeout = attr.value.float(6, inherit=True)

    def devices(self):
        """This function checks ADB to see if any devices are connected, if
        none are, it raises an exception, if there is at least one device
        connected:
        This function returns a list of tuples, each tuple contains the
        device id and device type. i.e. [['f0593056', 'device']] represents
        one connected device with id f0593056.
        """
        # with self.no_state_arguments:
        devices = (
            self.run('devices', pipe=True, check_stderr=True, check_return=True)
            .strip()
            .rstrip()
            .splitlines()[1:]
        )
        if len(devices) > 0:
            # At least one device found, lets return it in a nice way
            for index in range(len(devices)):
                devices[index] = devices[index].decode('utf-8').split('\t')
            return devices
        else:
            raise Exception('No devices found. Is the UE properly connected?')

    def is_device_connected(self, serialNum):
        """Uses the devices function to check if a device (sepecified by the
        serialNum argument) is connected to the ADB server, return true/false
        """
        # Get list of devices
        devices = self.devices()

        # Test to see if serialNum is one of the first elements in the list of tuples
        if serialNum not in [i[0] for i in devices]:
            return False
        return True

    def reboot(self, deviceId):
        """This function takes in a UE's Id, (from self.devices), and reboots
        the specified device.
        """
        # with self.no_state_arguments:
        if self.is_device_connected(deviceId):
            # Device is connected and ready to reboot, lets do
            self.run('-s', str(deviceId), 'reboot', check_return=True)
        else:
            # Devices isn't connected, lets raise an error saying so
            raise Exception('The specified device is not connected to the ADB server')

    def check_airplane_mode(self, deviceId):
        """Returns the status of airplane mode of device deviceId,
        returns 0 if deviceId's airplane_mode is OFF
        returns 1 if deviceId's airplane_mode is ON
        raises exception if it cannot find the device
        """
        if self.is_device_connected(deviceId):
            res = self.run(
                '-s',
                str(deviceId),
                'shell',
                'settings',
                'get',
                'global',
                'airplane_mode_on',
                pipe=True,
            )
            return res.strip().rstrip().decode('utf-8')
        else:
            raise Exception('The specified device is not connected to the ADB server')

    def set_airplane_mode(self, deviceId, status):
        """Sets the airplane_mode feature on the device specified by deviceId
        to the value of the 'status' argumentself.
        status should be set to either 0 or 1, 0 indicating that the airplane_mode
        should be turned off, 1 indicating it should be turned on
        """
        if self.is_device_connected(deviceId):
            if status not in [0, 1]:
                # invalid status argument
                raise Exception(
                    'The Airplane Mode feature can only be set to a value of 0 or 1'
                )
            else:
                self.foreground(
                    '-s',
                    str(deviceId),
                    'shell',
                    'settings',
                    'put',
                    'global',
                    'airplane_mode_on',
                    str(status),
                    pipe=True,
                    check_return=True,
                )
                self.foreground(
                    '-s',
                    str(deviceId),
                    'shell',
                    'am',
                    'broadcast',
                    '-a',
                    'android.intent.action.AIRPLANE_MODE',
                    pipe=True,
                    check_return=True,
                )
        else:
            raise Exception('The specified device is not connected to the ADB server')

    def push_file(self, deviceId, local_filepath, device_filepath):
        """Takes a file at the location specified by local_filepath and copys
        it into the directory specified by device_filepath on the device that
        is specified by deviceId.
        """
        if self.is_device_connected(deviceId):
            res = self.run(
                '-s',
                str(deviceId),
                'push',
                local_filepath,
                device_filepath,
                pipe=True,
                check_return=True,
            )
            res = res.strip().rstrip().splitlines()
            for val in res:
                print(val.decode('utf-8'))
        else:
            raise Exception('The specified device is not connected to the ADB server')


if __name__ == '__main__':
    with AndroidDebugBridge() as adb:
        devices = adb.devices()[0][0]  # Returns a list of tuples containing device Ids
        #  adb.reboot(devices[0][0])
        print(adb.check_airplane_mode(devices))
        adb.set_airplane_mode(devices, 0)
        print(adb.check_airplane_mode(devices))
        adb.push_file(
            devices,
            ssmdevices.lib.path('android', 'iperf'),
            '/data/local/tmp/iperf',
        )



__all__ = ['MiniCircuitsUSBSwitch']

from labbench import paramattr as attr

if __name__ == '__main__':
    # allow relative imports for the __main__ block below
    from _minicircuits_usb import SwitchAttenuatorBase
else:
    from ._minicircuits_usb import SwitchAttenuatorBase

__all__ = ['MiniCircuitsUSBSwitch']


class MiniCircuitsUSBSwitch(SwitchAttenuatorBase):
    # Mini-Circuits USB-SP4T-63
    _PID = 0x22

    @attr.property.int(min=1, max=4)
    def port(self):
        """the RF port connected to the COM port"""
        CMD_GET_SWITCH_PORT = 15
        d = self._cmd(CMD_GET_SWITCH_PORT)
        port = d[1]
        return port

    @port.setter
    def _(self, port):
        """the RF port connected to the COM port"""
        if port not in (1, 2, 3, 4):
            raise ValueError('Invalid switch port: %s' % port)
        self._cmd(port)



__all__ = ['WLANInfo', 'WLANClient']

import logging
import re
import subprocess as sp
import time

import labbench as lb
from labbench import paramattr as attr
import psutil

if __name__ == '__main__':
    from _networking import network_interface_info
else:
    from ._networking import network_interface_info


class WLANInfo(lb.ShellBackend):
    """parse calls to netsh to get information about WLAN interfaces on the host"""

    background_timeout = attr.value.float(5, inherit=True)
    binary_path = attr.value.Path(r'C:\Windows\System32\netsh.exe', must_exist=True)

    only_bssid: bool = attr.value.bool(False, key='mode=bsside', help='gather only BSSID information')
    interface: str = attr.value.str(None, key='interface', help='name of the interface to query')

    def wait(self):
        try:
            while self.running():
                pass
        except BaseException:
            pass

    @lb.retry(sp.TimeoutExpired, tries=5)
    def get_wlan_ssids(self, interface):
        def pairs_to_dict(lines):
            d = {}
            ssid = None

            for k, v in lines:
                k = k.lower().replace('(', '').replace(')', '').replace(' ', '_')
                if k.startswith('ssid'):
                    ssid = v
                    d[ssid] = {}
                elif ssid is not None:
                    d[ssid][k] = v
            return d

        # Execute the binary
        flags = lb.shell_options_from_keyed_values(self, hide_false=True, join_str='=')
        txt = self.run(
            self.binary_path, 'wlan', 'show', 'networks', *flags, check_return=True
        ).decode()

        # Parse into (key, value) pairs separated by whitespace and a colon
        lines = re.findall(r'^\s*(.*?)\s*:\s*(.*?)\s*$', txt, flags=re.MULTILINE)

        return pairs_to_dict(lines)

    @lb.retry(sp.TimeoutExpired, tries=5)
    def get_wlan_interfaces(self, name=None, param=None):
        def pairs_to_dict(lines):
            d = {}
            name = None

            for k, v in lines:
                k = k.lower().replace('(', '').replace(')', '').replace(' ', '_')
                if k.startswith('name'):
                    name = v
                    d[name] = {}
                elif name is not None:
                    if k == 'signal':
                        v = v.replace('%', '')
                    d[name][k] = v
            return d

        # Execute the binary
        flags = lb.shell_options_from_keyed_values(self, hide_false=True, join_str='=')
        txt = self.run(
            self.binary_path, 'wlan', 'show', 'interfaces', *flags, check_return=True
        ).decode()

        # Parse into (key, value) pairs separated by whitespace and a colon
        lines = re.findall(r'^\s*(\S+.*?)\s+:\s+(\S+.*?)\s*$', txt, flags=re.MULTILINE)

        return pairs_to_dict(lines)


class WLANClient(lb.Device):
    resource: str = attr.value.str(
        None,
        help='interface name (from the OS) or MAC address (nn:nn:nn:nn:nn)',
        cache=True,
    )
    ssid: str = attr.value.str(None, help='SSID of the AP for connection')
    timeout: float = attr.value.float(
        10,
        min=0,
        cache=True,
        help='attempt AP connection for this long before raising ConnectionError',
        label='s',
    )

    def open(self):
        self._import_pywifi()

        available = self.list_available_clients(by='physical_address')
        available_by_interface = {k: v['interface'] for k, v in available.items()}
        if (
            self.resource not in available_by_interface.keys()
            and self.resource not in available_by_interface.values()
        ):
            txt = (
                f"resource '{self.resource}' does not match any of the available "
                f'WLAN interface names {tuple(available_by_interface.keys())} '
                f'or corresponding MAC addresses {tuple(available_by_interface.values())}'
            )
            raise ConnectionError(txt)

        # Use netsh to identify the device guid from the network interface name
        info = network_interface_info(self.resource)
        mac = info['physical_address']

        if mac not in available:
            txt = (
                f"interface with MAC address '{self.resource}' is not one of the "
                f'WLAN interfaces {tuple(available)}'
            )
            print('raising connection error')
            raise ConnectionError(txt)

        guid = available[mac]['guid'].lower()

        ctrl = pywifi.wifi.wifiutil.WifiUtil()

        for iface_key in ctrl.interfaces():
            this_guid = str(iface_key['guid'])[1:-1].lower()
            if this_guid == guid:
                self.backend = pywifi.iface.Interface(iface_key)
                break
        else:
            # This really shouldn't happen
            raise ConnectionError('requested guid not present in pywifi')

        self._logger.debug(
            f"client network interface is '{info['interface']}' "
            f"at physical address '{info['physical_address']}'"
        )

    @classmethod
    def list_available_clients(cls, by='interface'):
        if by not in ('interface', 'guid', 'physical_address'):
            raise ValueError(
                f"argument 'by' must be one of ('interface', 'guid', 'physical_address'), not {by}"
            )

        netsh = WLANInfo()
        # netsh._logger.logger.disabled = True
        with netsh:
            # Check that this interface exists
            interfaces = netsh.get_wlan_interfaces()

        if by == 'interface':
            return interfaces

        ret = {}
        for if_name, if_map in interfaces.items():
            if_map['interface'] = if_name
            ret[if_map[by]] = if_map

        return ret

    @classmethod
    def _import_pywifi(cls):
        global pywifi

        # pywifi wants to clobber the global logging display settings with its own.
        # temporarily monkeypatch logging.basicConfig to bypass this
        try:
            logging.basicConfig, orig_config = lambda **kws: None, logging.basicConfig
            # level = lb.logger.logger.level
            import pywifi

        finally:
            logging.basicConfig = orig_config

        # reduce pywifi logging
        logger = logging.getLogger('pywifi')
        # logger.propagate = False
        # logger.disabled = True
        logger.setLevel(10000000000)

        cls._status_lookup = {
            pywifi.const.IFACE_CONNECTED: 'connected',
            pywifi.const.IFACE_CONNECTING: 'connecting',
            pywifi.const.IFACE_DISCONNECTED: 'disconnected',
            pywifi.const.IFACE_INACTIVE: 'inactive',
            pywifi.const.IFACE_SCANNING: 'scanning',
        }

    def interface_connect(self):
        if self.state == 'connected':
            return 0.0

        # Do the connect
        profile = pywifi.Profile()
        profile.ssid = self.ssid
        self.backend.connect(profile)

        t0 = time.perf_counter()
        while time.perf_counter() - t0 < self.timeout:
            if self.isup:
                break
            lb.sleep(0.02)
        else:
            raise TimeoutError(f'in connecting {repr(self)} to SSID {repr(self.ssid)}')

        t1 = time.perf_counter()
        while time.perf_counter() - t1 < self.timeout:
            s = self.state
            if s == 'connected':
                break
            lb.sleep(0.05)
        else:
            self._logger.debug(f'failed to connect to AP with SSID {repr(self.ssid)}')
            raise TimeoutError(
                'tried to connect but only achieved the {} state '.format(repr(s))
            )

        t2 = time.perf_counter()
        while time.perf_counter() - t2 < self.timeout:
            if self.channel is not None and self.signal is not None:
                break
            lb.sleep(0.02)
        else:
            raise TimeoutError('tried to connect, but got no AP scan information')

        time_elapsed = time.perf_counter() - t0
        self._logger.debug('connected WLAN interface to {}'.format(self.ssid))

        self.backend.scan()

        return time_elapsed

    def interface_disconnect(self):
        """Try to disconnect to the WLAN interface, or raise TimeoutError
        if there is no connection after the specified timeout.

        :param timeout: timeout to wait before raising TimeoutError
        :type timeout: float
        """
        if self.state == 'disconnected':
            return 0.0

        # Disconnect, if necessary
        self.backend.disconnect()

        # First, poll the faster state.isup
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < self.timeout:
            if not self.isup:
                break
            lb.sleep(0.02)
        else:
            raise TimeoutError('tried to disconnect but interface did not go down')

        # Then confirm with
        while time.perf_counter() - t0 < self.timeout:
            s = self.state
            if s == 'disconnected':
                break
            lb.sleep(0.05)
        else:
            raise TimeoutError(
                'tried to disconnect but only achieved the {} state '.format(repr(s))
            )

        self._logger.debug('disconnected WLAN interface')

    def interface_reconnect(self):
        """Reconnect to the network interface.

        :return: time elapsed to reconnect
        """
        self.interface_disconnect()
        return self.interface_connect()

    @attr.property.str(sets=False)
    def state(self):
        """`True` if psutil reports that the interface is up"""
        return self._status_lookup[self.backend.status()]

    @attr.property.bool(sets=False)
    def isup(self):
        """`True` if psutil reports that the interface is up"""
        stats = psutil.net_if_stats()
        iface_name = network_interface_info(self.resource)['interface']
        return stats[iface_name].isup

    @attr.property.int(sets=False, allow_none=True)
    def transmit_rate_mbps(self):
        stats = psutil.net_if_stats()
        iface_name = network_interface_info(self.resource)['interface']
        return stats[iface_name].speed

    @attr.property.int(allow_none=True, max=100, sets=False)
    def signal(self):
        def attempt():
            for result in self.backend.scan_results():
                if result.ssid == self.ssid:
                    return float(result.signal)
            else:
                lb.sleep(0.2)
                raise TimeoutError('interface reported no signal strength')

        t0 = time.perf_counter()
        while time.perf_counter() - t0 < 1.0:
            if self.state != 'scanning':
                break
            lb.sleep(0.02)
        else:
            raise TimeoutError('timeout while scanning for ssid signal strength')

        return lb.until_timeout(TimeoutError, 2 * self.timeout)(attempt)()

    @attr.property.str(sets=False, cache=True)
    def description(self):
        return self.backend.name()

    @attr.property.int(allow_none=True)
    def channel(self):
        def attempt():
            for result in self.backend.scan_results():
                if result.ssid == self.ssid:
                    return float(result.freq) * 1000
            else:
                lb.sleep(0.2)
                raise TimeoutError('interface reported no channel frequency')

        t0 = time.perf_counter()
        while time.perf_counter() - t0 < 1.0:
            if self.state != 'scanning':
                break
            lb.sleep(0.02)
        else:
            raise TimeoutError('ssid signal strength scan timeout')

        return lb.until_timeout(TimeoutError, 2 * self.timeout)(attempt)()

    def refresh(self):
        for attr_def in self._property_attrs:
            getattr(self, attr_def)


if __name__ == '__main__':
    client = WLANClient(
        resource='f8:ac:65:c8:72:bf',  # MAC address of the wireless client device on the host
        ssid='EnGenius1',  # SSID name to connect with
        timeout=5,  # (s) how long to continue attempts to connect the client and AP
    )

    with client:
        pass

    # with Netsh() as netsh:
    #     # Check that this interface exists
    #     available_interfaces = netsh.get_wlan_interfaces()
    #     print(available_interfaces)
    #
    #     print(netsh.get_wlan_ssids(list(available_interfaces)[0]))

    # with WLANStatus(resource='WLAN_Client_DUT', ssid='EnGenius1') as wlan:
    #     wlan.interface_reconnect()
    #     for attr in wlan._traits:
    #         print(attr, ':', getattr(wlan.state, attr))

#        while True:
#            print('isup: ', wlan.isup)
#            state = wlan.state
#            if state == 'connected':
#                try:
#                    print('RSSI "{}%"'.format(wlan.signal))
#                except:
#                    pass
#            else:
#                print(f'not connected - {state} instead')
#            lb.sleep(.25)



# -*- coding: utf-8 -*-
"""
Author: Keith Forsyth

Instrument driver for CMW500 base station emulator, initially created with intent of controlling band, 
power, and scheduling dynamics for UE IQ UL acquisitions.
"""

import time

import labbench as lb
from labbench import paramattr as attr

__all__ = ['RohdeSchwarzCMW500']

default_cmw_address = 'TCPIP0::10.0.0.3::inst0::INSTR'


@attr.visa_keying(remap={False: 'OFF', True: 'ON'})
class RohdeSchwarzCMW500(lb.VISADevice):
    """A base station emulator"""

    UPDATE_RATE = 20

    # Enhancement would be to change the SIGN1 to be a variable of some sort, and potentially the PCC to be a variable(SCC)
    channel_bandwidth = attr.method.str(
        key='CONF:LTE:SIGN1:CELL:BAND:PCC:DL',
        only=['B014', 'B030', 'B050', 'B100', 'B150', 'B200'],
    )
    lte_signaling = attr.method.bool(
        key='SOUR:LTE:SIGN1:CELL:STAT',
    )
    scheduling_type = attr.method.str(
        key='CONF:LTE:SIGN1:CONN:PCC:STYPE', only=['RMC', 'UDCH', 'UDTT']
    )  # not comprehensive

    # channel_band             = attr.method.str(key=NEED TO FIND COMMAND,values =[NEED TO FIND BAND CODES'OB66'])
    ue_attached = attr.method.bool(key='SENS:LTE:SIGN1:RRCS?')
    ulrmc_num_rbs = attr.method.int(key='CONF:LTE:SIGN1:CONN:PCC:RMCUL', min=1, max=100)
    ulrmc_modulation = attr.method.str(
        key='CONF:LTE:SIGN1:CONN:PCC:RMC:UL', only=['QPSK', 'Q16']
    )
    ulrmc_transblocksize = attr.method.int(
        key='CONF:LTE:SIGN1:CONN:PCC:RMC:UL', min=-1, max=100
    )
    ul_chan_freq = attr.method.float(key='CONF:LTE:SIGN1:RFS:PCC:CHAN:UL', sets=False)
    ue_target_power = attr.method.float(
        key='CONF:LTE:SIGN1:UL:PCC:PUSC:TPC:CLTP', min=-50, max=-20
    )
    ue_rsrp = attr.method.float(key='SENS:LTE:SIGN1:UER:PCC:RSRP?', sets=False)
    ue_rsrq = attr.method.float(key='SENS:LTE:SIGN1:UER:PCC:RSRQ?', sets=False)
    # spectrogram_num_avg      = attr.method.int(key='NEED TO FIND COMMAND',min=1,max=100)

    @ul_chan_freq.getter
    def _(self):
        command = 'CONF:LTE:SIGN1:RFS:PCC:CHAN:UL? Hz'
        return self.query(command)

    @ulrmc_num_rbs.getter
    def _(self):
        msg = self.query('CONF:LTE:SIGN1:CONN:PCC:RMC:UL?')
        value = msg.split(',')[0]
        num_rbs = value.split('N')[1]
        return num_rbs

    @ulrmc_modulation.getter
    def _(self):
        msg = self.query('CONF:LTE:SIGN1:CONN:PCC:RMC:UL?')
        modulation = msg.split(',')[1]
        return modulation

    @ulrmc_transblocksize.getter
    def _(self):
        msg = self.query('CONF:LTE:SIGN1:CONN:PCC:RMC:UL?')
        value = msg.split(',')[2]
        if value == 'KEEP':
            return -1
        if value == 'ZERO':
            return 0
        value = value.split('T')[1]
        return value

    @ulrmc_num_rbs.setter
    def _(self, value):
        modulation = self.state.ulrmc_modulation
        transblocksize = self.state.ulrmc_transblocksize
        if transblocksize == -1:
            transblocksize = 'KEEP'
        if transblocksize == 0:
            transblocksize = 'ZERO'
        if isinstance(transblocksize, int):
            transblocksize = 'T' + str(transblocksize)
        arg = f'N{value},{modulation},{transblocksize}'
        lb.logger.debug(f'ULRMC arg is: {arg}')
        self.write(f'CONF:LTE:SIGN1:CONN:PCC:RMC:UL {arg}')

    @ulrmc_modulation.setter
    def _(self, value):
        num_rbs = self.state.ulrmc_num_rbs
        transblocksize = self.state.ulrmc_transblocksize
        if transblocksize == -1:
            transblocksize = 'KEEP'
        if transblocksize == 0:
            transblocksize = 'ZERO'
        if isinstance(transblocksize, int):
            transblocksize = 'T' + str(transblocksize)
        arg = f'N{num_rbs},{value},{transblocksize}'
        lb.logger.debug(f'ULRMC arg is: {arg}')
        self.write(f'CONF:LTE:SIGN1:CONN:PCC:RMC:UL {arg}')

    @ulrmc_transblocksize.setter
    def _(self, value):
        num_rbs = self.state.ulrmc_num_rbs
        modulation = self.state.ulrmc_modulation
        if value == -1:
            value = 'KEEP'
        if value == 0:
            value = 'ZERO'
        if isinstance(value, int):
            value = 'T' + str(value)
        arg = f'N{num_rbs},{modulation},{value}'
        lb.logger.debug(f'ULRMC arg is: {arg}')
        self.write(f'CONF:LTE:SIGN1:CONN:PCC:RMC:UL {arg}')

    def wait_for_cell_activate(self, timeout=180):
        """a blocking wait until the cell has been stood up,checking the packet
        switched state as indicated by the CMW500
        SCPI = FETC:LTE:SIGN1:PSW:STAT?"""
        for i in range(timeout):
            cell_state = self.query('FETC:LTE:SIGN1:PSW:STAT?')
            packet_switched = None
            if cell_state == 'ATT':
                packet_switched = True
                break
            if cell_state == 'OFF':
                packet_switched = False
            time.sleep(1)
            if i % self.UPDATE_RATE == 0:
                lb.logger.info('Waiting for Cell to activate...')
        lb.logger.info('Packet Switching Activated: ' + str(packet_switched))
        return packet_switched

    def wait_for_cell_deactivate(self, timeout=180):
        """a blocking wait until the cell has been dekeyed,checking the packet
        switched state as indicated by the CMW500
        SCPI = FETCH:LTE:SIGN1:PSW:STAT?"""
        for i in range(timeout):
            cell_state = self.query('FETC:LTE:SIGN1:PSW:STAT?')
            packet_switched = None
            if cell_state == 'OFF':
                packet_switched = False
                break
            if cell_state == 'ATT':
                packet_switched = True
            time.sleep(1)
            if i % self.UPDATE_RATE == 0:
                lb.logger.info('Waiting for Cell to deactivate...')
        lb.logger.info('Packet Switching Activated: ' + str(packet_switched))
        return packet_switched

    def wait_for_ue_to_attach(self, timeout=180):
        """a blocking wait until the UE has attached,checking the RRC
        state as indicated by the CMW500
        SCPI = SENS:LTE:SIGN1:RRCS?"""
        for i in range(timeout):
            ue_state = self.query('SENS:LTE:SIGN1:RRCS?')
            attached = None
            if ue_state == 'CONN':
                attached = True
                break
            if ue_state == 'IDLE':
                attached = False
            time.sleep(1)
            if i % self.UPDATE_RATE == 0:
                lb.logger.info('Waiting for UE to attach...')
        lb.logger.info('UE is Attached: ' + str(attached))
        return attached

    def fetch_spectrogram(self):
        """a request to pull the spectrogram down from the cmw500, input
        values are fixed with a RBW of 30kHz and the frequency span x array
        calculated on that assumption.  Would be better to generalize this.
        May also want to add the ability to trigger the measurement
        SCPI = FETC:LTE:MEAS:MEV:TRAC:SEM:RBW:AVER?"""
        values = self.query_ascii_values('FETC:LTE:MEAS:MEV:TRAC:SEM:RBW30:AVER?')
        cent_freq = self.state.ul_chan_freq
        values.pop(0)
        freq_from_cent_list = []
        abs_freq_list = []
        for i in range(len(values)):
            abs_freq = cent_freq + (i - (len(values) - 1) / 2) * 30000
            freq_from_cent = (i - (len(values) - 1) / 2) * 30000
            freq_from_cent_list.append(freq_from_cent)
            abs_freq_list.append(abs_freq)
        return {
            'abs_freq': abs_freq_list,
            'freq_from_cent': freq_from_cent_list,
            'power': values,
        }

    def fetch_iq(self):
        """a request to pull the iq down from the cmw500, not implemented yet"""
        raise NotImplementedError

    def config_ul_udtti(self, tti, start_rb, num_rb, mod):
        """Configure the CMW for User Defined TTI scheduling and send
        start rb,number of rb's and type of modulation to CMW for a given TTI.
        Available modulations for UL are QPSK and 16QAM (Q16).
        SCPI = CONF:LTE:SIGN1:CONN:PCC:UDTT:UL {tti},{start_rb},{num_rb},{mod}"""
        self.state.scheduling_type = 'UDTT'
        self.write(f'CONF:LTE:SIGN1:CONN:PCC:UDTT:UL {tti},{start_rb},{num_rb},{mod},0')
        ul_udtti = self.query(f'CONF:LTE:SIGN1:CONN:PCC:UDTT:UL? {tti}').split(',')
        if int(ul_udtti[0]) != start_rb:
            self._logger.error('ul_udtti write failed')
            return False
        if int(ul_udtti[1]) != int(num_rb):
            self._logger.error('ul_udtti write failed')
            return False
        if ul_udtti[2] != str(mod):
            self._logger.error('ul_udtti write failed')
            return False
        self._logger.debug(
            'ul_udtti written, tti:{tti},start_rb:{start_rb},num_rb:{num_rb},mod:{mod}'
        )
        return True


def test_cmw_idn():
    """Use the IDN query to check that the manf and model # match the expected"""
    cmw500 = RohdeSchwarzCMW500(default_cmw_address)
    cmw500.connect()
    cmw_identity = cmw500.query('*IDN?').split(',')
    if cmw_identity[0] == 'Rohde&Schwarz' and cmw_identity[1] == 'CMW':
        return True
    else:
        return False


def test_cmw():
    """Test to:
        1.  connect to the instrument
        2.  set the bandwidth
        3.  set UE target power
        4.  set RMC #RBs
        5.  stand the cell up
        6.  wait for the UE to attach
        7.  shut the cell down.
    This requires the UE to be in a state that in can connect with the instrument"""
    cmw500 = RohdeSchwarzCMW500(default_cmw_address)
    cmw500.connect()
    cmw500.state.channel_bandwidth = 'B014'
    # cmw500.state.channel_band = 'NEED VALUE'
    cmw500.state.ue_target_power = -21
    cmw500.state.ulrmc_num_rbs = 5
    cmw500.state.lte_signaling = True
    cmw500.wait_for_ue_to_attach(timeout=60)
    cmw500.state.lte_signaling = False
    cmw500.wait_for_cell_deactivate(timeout=60)
    return True


if __name__ == '__main__':
    print('IDN test pass: ' + str(test_cmw_idn()))
    print('UE Attach test pass: ' + str(test_cmw()))



__all__ = ['MiniCircuitsRCDAT']

import time
import labbench as lb
from labbench import paramattr as attr

if __name__ == '__main__':
    from _minicircuits_usb import SwitchAttenuatorBase
else:
    from ._minicircuits_usb import SwitchAttenuatorBase


class MiniCircuitsRCDAT(SwitchAttenuatorBase):
    # SwitchAttenuatorBase uses product ID to connect to USB devices
    _PID = 0x23

    frequency: float = attr.value.float(
        default=None,
        allow_none=True,
        min=10e6,
        max=6e9,
        help='frequency for calibration data (None for no calibration)',
        label='Hz',
    )

    output_power_offset: float = attr.value.float(
        default=None,
        allow_none=True,
        help='output power level at 0 dB attenuation',
        label='dBm',
    )

    calibration_path: float = attr.value.Path(
        default=None,
        allow_none=True,
        cache=True,
        must_exist=True,
        help='path to the calibration table csv file (containing frequency '
        '(row) and attenuation setting (column)), or None to search ssmdevices',
    )

    channel: int = attr.value.int(
        default=1,
        allow_none=True,
        min=1,
        max=4,
        cache=True,
        help='a port selector for 4 port attenuators None is a single attenuator',
    )

    # this is the only property that directly sets attenuation in the device
    @attr.property.float(
        min=0, max=115, step=0.25, label='dB', help='uncalibrated attenuation'
    )
    def attenuation_setting(self):
        # getter
        CMD_GET_ATTENUATION = 18

        if self.channel is None:
            d = self._cmd(CMD_GET_ATTENUATION)
            full_part = d[1]
            frac_part = float(d[2]) / 4.0
            return full_part + frac_part
        elif self.channel in range(1, 5):
            d = self._cmd(CMD_GET_ATTENUATION)
            offs = self.channel * 2 - 1
            full_part = d[offs]
            frac_part = float(d[offs + 1]) / 4.0
            return full_part + frac_part
        else:
            raise AttributeError

    @attenuation_setting.setter
    def _(self, set_value):
        # setter
        CMD_SET_ATTENUATION = 19

        if self.channel is None:
            value1 = int(set_value)
            value2 = int((set_value - value1) * 4.0)
            self._cmd(CMD_SET_ATTENUATION, value1, value2, 1)
        elif self.channel in range(1, 5):
            value1 = int(set_value)
            value2 = int((set_value - value1) * 4.0)
            self._cmd(CMD_SET_ATTENUATION, value1, value2, self.channel)
            time.sleep(0.001)  # this prevents collisions for the moment
        else:
            raise AttributeError

    # the remaining traits are calibration corrections for attenuation_setting
    attenuation = attenuation_setting.corrected_from_table(
        allow_none=True,
        path_attr=calibration_path,
        index_lookup_attr=frequency,
        table_index_column='Frequency(Hz)',
        help='calibrated attenuation',
    )

    output_power = attenuation.corrected_from_expression(
        -attenuation + output_power_offset,
        help='calibrated output power level',
        label='dBm',
    )


if __name__ == '__main__':
    import numpy as np

    check_single_channel = True
    check_four_channel = False
    #    resource = '12208250156'
    resource = '12104060052'

    lb.show_messages('info')

    if check_single_channel:
        for i in np.arange(0, 110.25, 5):
            atten = MiniCircuitsRCDAT(resource, frequency=5.3e9)
            with atten:
                atten.attenuation_setting = i
                print(f'Attenuator set point {str(atten.attenuation_setting)}')

    if check_four_channel:
        for i in np.arange(0, 110.25, 5):
            atten1 = MiniCircuitsRCDAT(resource, frequency=5.3e9, channel=1)
            atten2 = MiniCircuitsRCDAT(resource, frequency=5.3e9, channel=2)
            atten3 = MiniCircuitsRCDAT(resource, frequency=5.3e9, channel=3)
            atten4 = MiniCircuitsRCDAT(resource, frequency=5.3e9, channel=4)
            with atten1, atten2, atten3, atten4:
                atten1.attenuation_setting = i
                atten2.attenuation_setting = i
                atten3.attenuation_setting = i
                atten4.attenuation_setting = i
                print(f'Attenuator 1 set point {str(atten1.attenuation_setting)}')
                print(f'Attenuator 2 set point {str(atten2.attenuation_setting)}')
                print(f'Attenuator 3 set point {str(atten3.attenuation_setting)}')
                print(f'Attenuator 4 set point {str(atten4.attenuation_setting)}')
